#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass

from eth_typing import ChecksumAddress

from skale.types.node import NodeId


@dataclass
class ExitRequest:
    request_id: int
    user: ChecksumAddress
    node_id: NodeId
    amount: int
    unlock_date: int

    def to_dict(self) -> dict:
        return {
            'request_id': self.request_id,
            'user': self.user,
            'node_id': self.node_id,
            'amount': self.amount,
            'unlock_date': self.unlock_date,
        }


def exit_request_from_tuple(data: tuple) -> ExitRequest:
    return ExitRequest(
        request_id=data[0],
        user=data[1],
        node_id=data[2],
        amount=data[3],
        unlock_date=data[4],
    )



from tests.constants import DEFAULT_SCHAIN_HASH, DEFAULT_SCHAIN_NAME


def test_get_previous_groups_public_key(skale):
    group_id = skale.web3.keccak(text=DEFAULT_SCHAIN_NAME)
    public_key = skale.key_storage.get_previous_public_key(group_id)
    assert isinstance(public_key, tuple)
    assert len(public_key) == 2
    assert isinstance(public_key[0], tuple) and len(public_key[0]) == 2
    assert isinstance(public_key[1], tuple) and len(public_key[1]) == 2


def test_get_all_previous_groups_public_keys(skale):
    # TODO: Improve test
    group_id = skale.web3.keccak(text=DEFAULT_SCHAIN_NAME)
    keys = skale.key_storage.get_all_previous_public_keys(group_id)
    assert keys == []


def test_get_group_public_key(skale):
    assert skale.key_storage.get_common_public_key(DEFAULT_SCHAIN_HASH) == ((0, 0), (0, 0))



"""Provision Allocator contracts for testing"""

from skale import SkaleAllocator, SkaleManager
from skale.utils.contracts_provision.allocator import add_test_plan, transfer_tokens_to_allocator
from skale.utils.contracts_provision.main import add_test_permissions, setup_validator
from skale.utils.helper import get_allocator_address, get_skale_manager_address, init_default_logger
from skale.utils.web3_utils import init_web3
from skale.wallets import Web3Wallet
from tests.constants import (
    ENDPOINT,
    ETH_PRIVATE_KEY,
    TEST_ABI_FILEPATH,
    TEST_ALLOCATOR_ABI_FILEPATH,
)


def init_libs():
    web3 = init_web3(ENDPOINT)
    wallet = Web3Wallet(ETH_PRIVATE_KEY, web3)
    return (
        SkaleManager(ENDPOINT, get_skale_manager_address(TEST_ABI_FILEPATH), wallet),
        SkaleAllocator(ENDPOINT, get_allocator_address(TEST_ALLOCATOR_ABI_FILEPATH), wallet),
    )


def provision_contracts():
    init_default_logger()
    skale_manager, skale_allocator = init_libs()
    add_test_permissions(skale_manager)
    setup_validator(skale_manager)

    launch_ts = skale_manager.constants_holder.get_launch_timestamp()
    if launch_ts != 0:
        skale_manager.constants_holder.set_launch_timestamp(0, wait_for=True)

    vesting_manager_role = skale_allocator.allocator.vesting_manager_role()
    skale_allocator.allocator.grant_role(vesting_manager_role, skale_allocator.wallet.address)
    transfer_tokens_to_allocator(skale_manager, skale_allocator)
    add_test_plan(skale_allocator)


if __name__ == '__main__':
    provision_contracts()



import pytest

from tests.constants import MONTH_IN_SECONDS


def test_get_current_month(skale):
    current_month = skale.time_helpers_with_debug.get_current_month()
    assert isinstance(current_month, int)
    assert 0 < current_month < 1200


@pytest.mark.skip('Not working. Fix later')
def test_skip_time(skale):
    current_month_before = skale.time_helpers_with_debug.get_current_month()
    skale.time_helpers_with_debug.skip_time(MONTH_IN_SECONDS + 1, wait_for=True)
    current_month_after = skale.time_helpers_with_debug.get_current_month()
    assert current_month_after == current_month_before + 1, (
        current_month_after,
        current_month_before,
    )



# flake8: noqa
from skale.contracts.skale_contract import SkaleContract
from skale.contracts.base_contract import transaction_method
from skale.contracts.paymaster.paymaster import Paymaster

__all__ = ['Paymaster']



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
"""SKALE Allocator Core Escrow methods"""

import functools
from typing import Any, Callable

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction
from web3.types import Wei

from skale.contracts.allocator.allocator import Allocator
from skale.contracts.allocator_contract import AllocatorContract
from skale.contracts.base_contract import transaction_method
from skale.transactions.result import TxRes
from skale.types.delegation import DelegationId
from skale.types.validator import ValidatorId


def beneficiary_escrow(transaction: Callable[..., TxRes]) -> Callable[..., TxRes]:
    @functools.wraps(transaction)
    def wrapper(
        self: AllocatorContract, *args: Any, beneficiary_address: ChecksumAddress, **kwargs: Any
    ) -> TxRes:
        self.contract = self.skale.instance.get_contract('Escrow', beneficiary_address)
        return transaction(self, *args, **kwargs)

    return wrapper


class Escrow(AllocatorContract):
    @property
    @functools.lru_cache()
    def allocator(self) -> Allocator:
        return self.skale.allocator

    @beneficiary_escrow
    @transaction_method
    def retrieve(self) -> 'ContractFunction':
        """Allows Holder to retrieve vested tokens from the Escrow contract

        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.retrieve()

    @beneficiary_escrow
    @transaction_method
    def retrieve_after_termination(self, address: ChecksumAddress) -> 'ContractFunction':
        """Allows Core Owner to retrieve remaining transferrable escrow balance
        after Core holder termination. Slashed tokens are non-transferable

        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.retrieveAfterTermination(address)

    @beneficiary_escrow
    @transaction_method
    def delegate(
        self, validator_id: ValidatorId, amount: Wei, delegation_period: int, info: str
    ) -> 'ContractFunction':
        """Allows Core holder to propose a delegation to a validator

        :param validator_id: ID of the validator to delegate tokens
        :type validator_id: int
        :param amount: Amount of tokens to delegate
        :type amount: int
        :param delegation_period: Period of delegation
        :type delegation_period: int
        :param info: Delegation request information
        :type info: str
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.delegate(validator_id, amount, delegation_period, info)

    @beneficiary_escrow
    @transaction_method
    def request_undelegation(self, delegation_id: DelegationId) -> 'ContractFunction':
        """Allows Holder and Owner to request undelegation. Only Owner can
        request undelegation after Core holder is deactivated (upon holder termination)

        :param delegation_id: ID of the delegation
        :type delegation_id: int
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.requestUndelegation(delegation_id)

    @beneficiary_escrow
    @transaction_method
    def withdraw_bounty(self, validator_id: ValidatorId, to: ChecksumAddress) -> 'ContractFunction':
        """Allows Beneficiary and Vesting Owner to withdraw earned bounty.

        :param validator_id: ID of the validator
        :type validator_id: int
        :param to: Recipient address
        :type to: str
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.withdrawBounty(validator_id, to)

    @beneficiary_escrow
    @transaction_method
    def cancel_pending_delegation(self, delegation_id: DelegationId) -> 'ContractFunction':
        """Cancel pending delegation request.

        :param delegation_id: ID of the delegation to cancel
        :type delegation_id: int
        :returns: Transaction results
        :rtype: TxRes
        """
        return self.contract.functions.cancelPendingDelegation(delegation_id)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.ima.schain.base_token_manager import BaseTokenManager
from skale.types.schain import SchainName
from skale.utils.helper import schain_name_to_hash


class TokenManagerERC20(BaseTokenManager):
    @transaction_method
    def exit_to_main_erc20(self, token_address: ChecksumAddress, amount: int) -> ContractFunction:
        return self.contract.functions.exitToMainERC20(token_address, amount)

    @transaction_method
    def transfer_to_schain_erc20(
        self, schain_name: SchainName, token_address: ChecksumAddress, amount: int
    ) -> ContractFunction:
        return self.contract.functions.transferToSchainERC20(schain_name, token_address, amount)

    @transaction_method
    def transfer_to_schain_erc20_direct(
        self, schain_name: SchainName, token_address: ChecksumAddress, amount: int, receiver=str
    ) -> ContractFunction:
        return self.contract.functions.transferToSchainERC20Direct(
            schain_name, token_address, amount, receiver
        )

    @transaction_method
    def transfer_to_schain_hash_erc20_direct(
        self, schain_hash: str, token_address: ChecksumAddress, amount: int, receiver=str
    ) -> ContractFunction:
        return self.contract.functions.transferToSchainHashERC20Direct(
            schain_hash, token_address, amount, receiver
        )

    @transaction_method
    def add_erc20_token(
        self,
        origin_chain_name: SchainName,
        origin_address: ChecksumAddress,
        clone_address: ChecksumAddress,
    ) -> ContractFunction:
        return self.contract.functions.addERC20TokenByOwner(
            origin_chain_name, origin_address, clone_address
        )

    def get_clone(self, schain_name: SchainName, origin_address: ChecksumAddress) -> int:
        return self.contract.functions.clonesErc20(
            schain_name_to_hash(schain_name), origin_address
        ).call()



"""SKALE config test"""

import logging
from unittest import mock

import pytest
from web3.auto import w3

from skale import SkaleManager
from skale.utils.account_tools import generate_account, send_eth
from skale.utils.contracts_provision.fake_multisig_contract import deploy_fake_multisig_contract
from skale.utils.contracts_provision.main import (
    add_test2_schain_type,
    add_test_permissions,
    cleanup_nodes,
    cleanup_schains,
    create_nodes,
    create_schain,
    link_nodes_to_validator,
    set_automining,
    set_default_mining_interval,
    set_mining_interval,
    setup_validator,
)
from skale.utils.contracts_provision.utils import generate_random_node_data
from skale.utils.helper import get_skale_manager_address
from skale.utils.web3_utils import init_web3
from skale.wallets import Web3Wallet
from tests.constants import ENDPOINT, TEST_ABI_FILEPATH
from tests.helper import init_fair, init_skale, init_skale_allocator

ETH_AMOUNT_PER_NODE = 1
logger = logging.getLogger(__name__)

_last_finished_test_name: str | None = None


@pytest.fixture(scope='session')
def web3():
    """Returns a SKALE Manager instance with provider from config"""
    w3 = init_web3(ENDPOINT)
    return w3


@pytest.fixture(scope='session')
def skale(web3, request):
    """Returns a cached SKALE Manager instance with provider from config"""
    if not hasattr(request.config, '_cached_skale'):
        skale_obj = init_skale(web3)
        add_test_permissions(skale_obj)
        add_test2_schain_type(skale_obj)
        if skale_obj.constants_holder.get_launch_timestamp() != 0:
            skale_obj.constants_holder.set_launch_timestamp(0)
        deploy_fake_multisig_contract(skale_obj.web3, skale_obj.wallet)
        request.config._cached_skale = skale_obj
    return request.config._cached_skale


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    report = outcome.get_result()
    if report.when == 'call' and report.outcome in {'passed', 'failed', 'skipped'}:
        item.config._last_finished_test_name = item.name


@pytest.fixture(scope='session')
def _skale_stats_logger(skale):
    def _log(test_name: str) -> None:
        logger.warning(
            test_name
            + ' '
            + str(skale.stats.http_posts)
            + ' '
            + str(skale.stats.by_method.most_common(10))
        )

    return _log


@pytest.fixture(autouse=True)
def log_prev_test_stats(request, _skale_stats_logger):
    global _last_finished_test_name
    prev_name = getattr(request.config, '_last_finished_test_name', None)
    if prev_name is not None:
        _skale_stats_logger(prev_name)
    _last_finished_test_name = request.node.name


@pytest.fixture(scope='session')
def fair(web3):
    return init_fair(web3)


@pytest.fixture(scope='session')
def validator(skale):
    return setup_validator(skale)


@pytest.fixture(scope='session')
def number_of_nodes():
    return 2


@pytest.fixture
def node_wallets(skale, number_of_nodes):
    wallets = []
    for i in range(number_of_nodes):
        acc = generate_account(skale.web3)
        pk = acc['private_key']
        wallet = Web3Wallet(pk, skale.web3)
        send_eth(
            web3=skale.web3,
            wallet=skale.wallet,
            receiver_address=wallet.address,
            amount=ETH_AMOUNT_PER_NODE,
        )
        wallets.append(wallet)
    return wallets


@pytest.fixture
def node_skales(skale, node_wallets):
    return [
        SkaleManager(ENDPOINT, get_skale_manager_address(TEST_ABI_FILEPATH), wallet)
        for wallet in node_wallets
    ]


@pytest.fixture
def nodes(skale, node_skales, validator):
    link_nodes_to_validator(skale, validator, node_skales)
    ids = create_nodes(node_skales)
    try:
        yield ids
    finally:
        cleanup_nodes(skale, ids)


@pytest.fixture
def fair_active_nodes(fair, node_wallets):
    main_wallet = fair.wallet

    for wallet in node_wallets:
        fair.wallet = wallet
        ip, _, port, _ = generate_random_node_data()
        self_stake_requirement = fair.staking.self_stake_requirement()
        fair.nodes.register_active(ip=ip, port=port, value=self_stake_requirement)

    fair.wallet = main_wallet
    try:
        yield node_wallets
    finally:
        """TODO: Remove the node from the fair instance."""


@pytest.fixture
def fair_passive_nodes(fair, node_wallets):
    main_wallet = fair.wallet

    for wallet in node_wallets:
        fair.wallet = wallet
        ip, _, port, _ = generate_random_node_data()
        fair.nodes.register_passive(ip=ip, port=port)

    fair.wallet = main_wallet
    try:
        yield node_wallets
    finally:
        """TODO: Remove the node from the fair instance."""


@pytest.fixture
def schain(skale, nodes):
    try:
        yield create_schain(
            skale,
            schain_type=1,  # test2 should have 1 index
            random_name=True,
        )
    finally:
        cleanup_schains(skale)


@pytest.fixture
def skale_allocator(web3):
    """Returns a SKALE Allocator instance with provider from config"""
    return init_skale_allocator(web3)


@pytest.fixture
def empty_account():
    return w3.eth.account.create()


@pytest.fixture
def failed_skale(skale):
    tmp_wait, tmp_sign_and_send = skale.wallet.wait, skale.wallet.sign_and_send
    skale.wallet.sign_and_send = mock.Mock(return_value='0x000000000')
    skale.wallet.wait = mock.Mock(return_value={'status': 0, 'error': 'Test error'})
    try:
        yield skale
    finally:
        skale.wallet.sign_and_send = tmp_sign_and_send
        skale.wallet.wait = tmp_wait


@pytest.fixture
def block_in_seconds(skale):
    # Mine block every three seconds without automine
    # Makes web3.py throw exception in a same way as for geth
    try:
        set_automining(skale.web3, False)
        set_mining_interval(skale.web3, 3)
        yield
    finally:
        set_automining(skale.web3, True)
        set_default_mining_interval(skale.web3)



import pytest


@pytest.mark.parametrize('number_of_nodes', [1])
def test_is_healthy(fair, fair_active_nodes):
    prev_heartbeat_interval = fair.status.contract.functions.heartbeatInterval().call()
    fair.status.set_heartbeat_interval(5 * 60)

    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    is_healthy = fair.status.is_healthy(node_id)
    assert isinstance(is_healthy, bool)
    assert is_healthy is False

    main_wallet = fair.wallet
    fair.wallet = fair_active_nodes[0]
    fair.status.alive()
    fair.wallet = main_wallet

    is_healthy_after_alive = fair.status.is_healthy(node_id)
    assert isinstance(is_healthy_after_alive, bool)
    assert is_healthy_after_alive is True

    fair.status.set_heartbeat_interval(prev_heartbeat_interval)


@pytest.mark.parametrize('number_of_nodes', [1])
def test_node_becomes_enabled_after_whitelist_and_heartbeat(fair, fair_active_nodes):
    prev_heartbeat_interval = fair.status.contract.functions.heartbeatInterval().call()
    fair.status.set_heartbeat_interval(5 * 60)

    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    assert fair.staking.is_node_enabled(node_id) is False
    fair.status.whitelist_node(node_id)
    assert fair.staking.is_node_enabled(node_id) is False

    fair.staking.stake(node_id, value=10000)

    main_wallet = fair.wallet
    fair.wallet = fair_active_nodes[0]
    fair.status.alive()
    fair.wallet = main_wallet

    assert fair.staking.is_node_enabled(node_id) is True

    fair.status.set_heartbeat_interval(prev_heartbeat_interval)


@pytest.mark.parametrize('number_of_nodes', [1])
def test_calc_alive_gas_limit(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id
    fair.status.whitelist_node(node_id)
    fair.staking.stake(node_id, value=10000)

    main_wallet = fair.wallet
    fair.wallet = fair_active_nodes[0]
    alive_gas_limit = fair.status.calc_alive_gas_limit()
    res = fair.status.alive(gas_limit=alive_gas_limit)
    fair.wallet = main_wallet

    tx = fair.web3.eth.get_transaction(res.tx_hash)
    assert tx['gas'] == alive_gas_limit


@pytest.mark.parametrize('number_of_nodes', [1])
def test_get_whitelisted_nodes(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id
    fair.status.whitelist_node(node_id)

    whitelisted_nodes = fair.status.get_whitelisted_nodes()
    assert isinstance(whitelisted_nodes, list)
    assert all(isinstance(node_id, int) for node_id in whitelisted_nodes)
    assert node_id in whitelisted_nodes


@pytest.mark.parametrize('number_of_nodes', [1])
def test_is_whitelisted(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    is_whitelisted = fair.status.is_whitelisted(node_id)
    assert isinstance(is_whitelisted, bool)
    assert is_whitelisted is False


@pytest.mark.parametrize('number_of_nodes', [1])
def test_alive(fair, fair_active_nodes):
    prev_heartbeat_interval = fair.status.contract.functions.heartbeatInterval().call()
    fair.status.set_heartbeat_interval(5 * 60)

    main_wallet = fair.wallet
    fair.wallet = fair_active_nodes[0]

    fair.status.alive()

    fair.wallet = main_wallet

    fair.status.set_heartbeat_interval(prev_heartbeat_interval)


def test_set_heartbeat_interval(fair):
    prev_heartbeat_interval = fair.status.contract.functions.heartbeatInterval().call()

    interval = 2025
    fair.status.set_heartbeat_interval(interval)

    new_heartbeat_interval = fair.status.contract.functions.heartbeatInterval().call()
    assert new_heartbeat_interval == interval

    fair.status.set_heartbeat_interval(prev_heartbeat_interval)


@pytest.mark.parametrize('number_of_nodes', [1])
def test_whitelist_node(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    initial_whitelisted = fair.status.is_whitelisted(node_id)
    assert initial_whitelisted is False

    fair.status.whitelist_node(node_id)
    is_whitelisted_after = fair.status.is_whitelisted(node_id)
    assert isinstance(is_whitelisted_after, bool)
    assert is_whitelisted_after is True


@pytest.mark.parametrize('number_of_nodes', [1])
def test_remove_node_from_whitelist(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    fair.status.whitelist_node(node_id)
    assert fair.status.is_whitelisted(node_id) is True

    fair.status.remove_node_from_whitelist(node_id)

    after_removal = fair.status.is_whitelisted(node_id)
    assert after_removal is False



"""Tests for contracts/delegation/delegation_controller.py"""

import pytest

from skale.contracts.manager.delegation.delegation_controller import FIELDS
from skale.transactions.exceptions import ContractLogicError
from skale.transactions.result import DryRunRevertError
from skale.types.delegation import DelegationStatus
from skale.utils.contracts_provision.main import _skip_evm_time
from skale.utils.contracts_provision.utils import generate_random_name
from tests.constants import (
    D_DELEGATION_AMOUNT,
    D_DELEGATION_INFO,
    D_DELEGATION_PERIOD,
    D_VALIDATOR_ID,
    DELEGATION_STRUCT_LEN,
    MONTH_IN_SECONDS,
    NOT_EXISTING_ID,
)


def _delegate_and_activate(skale, validator_id=D_VALIDATOR_ID):
    skale.delegation_controller.delegate(
        validator_id=validator_id,
        amount=D_DELEGATION_AMOUNT,
        delegation_period=D_DELEGATION_PERIOD,
        info=D_DELEGATION_INFO,
        wait_for=True,
    )
    delegations = skale.delegation_controller.get_all_delegations_by_validator(validator_id)
    skale.delegation_controller.accept_pending_delegation(delegations[-1]['id'], wait_for=True)
    _skip_evm_time(skale.web3, MONTH_IN_SECONDS)


def _get_number_of_delegations(skale, validator_id=D_VALIDATOR_ID):
    return skale.delegation_controller._get_delegation_ids_len_by_validator(validator_id)


def test_get_raw_not_exist(skale):
    with pytest.raises((ContractLogicError, ValueError)):
        skale.delegation_controller._DelegationController__raw_get_delegation(NOT_EXISTING_ID)


def test_get_delegation(skale, validator):
    validator_id = validator
    delegations_by_validator = skale.delegation_controller._get_delegation_ids_by_validator(
        validator_id
    )
    delegation_id = delegations_by_validator[-1]
    delegation_struct = skale.delegation_controller._DelegationController__raw_get_delegation(
        delegation_id
    )
    assert delegation_struct[1] == validator_id
    assert len(delegation_struct) == DELEGATION_STRUCT_LEN

    delegation = skale.delegation_controller.get_delegation(delegation_id)
    assert list(delegation.keys()) == FIELDS
    assert delegation['validator_id'] == validator_id
    assert [k for k, v in delegation.items() if v is None] == []


def test_delegate(skale, validator):
    validator_id = validator
    num_of_delegations_before = _get_number_of_delegations(skale)
    delegated_now_before = skale.delegation_controller.get_delegated_to_validator_now(validator_id)
    delegated_amount_before = skale.delegation_controller.get_delegated_amount(skale.wallet.address)

    _delegate_and_activate(skale)
    num_of_delegations_after = _get_number_of_delegations(skale)
    assert num_of_delegations_after == num_of_delegations_before + 1

    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=validator_id
    )
    assert delegations[-1]['info'] == D_DELEGATION_INFO

    delegated_now_after = skale.delegation_controller.get_delegated_to_validator_now(validator_id)
    delegated_amount_after = skale.delegation_controller.get_delegated_amount(skale.wallet.address)
    assert delegated_now_after == delegated_now_before + D_DELEGATION_AMOUNT
    assert delegated_amount_after == delegated_amount_before + D_DELEGATION_AMOUNT

    month = skale.time_helpers_with_debug.get_current_month()
    res = skale.delegation_controller.get_delegated_to_validator(validator_id, month)
    assert isinstance(res, int)


def test_get_delegation_ids_by_validator(skale, validator):
    validator_id = validator
    delegation_ids = skale.delegation_controller._get_delegation_ids_by_validator(
        validator_id=validator_id
    )
    delegation_ids_len = skale.delegation_controller._get_delegation_ids_len_by_validator(
        validator_id
    )
    assert len(delegation_ids) == delegation_ids_len
    latest_delegation = skale.delegation_controller.get_delegation(delegation_ids[-1])
    assert latest_delegation['validator_id'] == validator_id


def test_get_all_delegations_by_holder(skale):
    delegations = skale.delegation_controller.get_all_delegations_by_holder(skale.wallet.address)
    assert all([delegation['address'] == skale.wallet.address for delegation in delegations])


def test_get_all_delegations_by_validator(skale, validator):
    validator_id = validator
    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=validator_id
    )
    assert all([delegation['validator_id'] == validator_id for delegation in delegations])


def test_accept_pending_delegation(skale, validator):
    validator_id = validator
    info = f'{D_DELEGATION_INFO}-{generate_random_name()}'
    skale.delegation_controller.delegate(
        validator_id=validator_id,
        amount=D_DELEGATION_AMOUNT,
        delegation_period=D_DELEGATION_PERIOD,
        info=info,
        wait_for=True,
    )
    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=validator_id
    )
    delegation_id = delegations[-1]['id']
    assert delegations[-1]['status'] == DelegationStatus.PROPOSED
    assert delegations[-1]['info'] == info
    skale.delegation_controller.accept_pending_delegation(delegation_id)
    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=validator_id
    )
    assert delegations[-1]['id'] == delegation_id
    assert delegations[-1]['status'] == DelegationStatus.ACCEPTED
    assert delegations[-1]['info'] == info


def test_cancel_pending_delegation(skale, validator):
    validator_id = validator
    skale.delegation_controller.delegate(
        validator_id=validator_id,
        amount=D_DELEGATION_AMOUNT,
        delegation_period=D_DELEGATION_PERIOD,
        info=D_DELEGATION_INFO,
        wait_for=True,
    )
    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=validator_id
    )
    delegation_id = delegations[-1]['id']
    assert delegations[-1]['status'] == DelegationStatus.PROPOSED
    skale.delegation_controller.cancel_pending_delegation(delegation_id, wait_for=True)
    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=validator_id
    )
    assert delegations[-1]['id'] == delegation_id
    assert delegations[-1]['status'] == DelegationStatus.CANCELED


def test_request_undelegate(skale, validator):
    validator_id = validator
    skale.delegation_controller.delegate(
        validator_id=validator_id,
        amount=D_DELEGATION_AMOUNT,
        delegation_period=D_DELEGATION_PERIOD,
        info=D_DELEGATION_INFO,
        wait_for=True,
    )
    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=validator_id
    )
    delegation_id = delegations[-1]['id']
    skale.delegation_controller.accept_pending_delegation(delegations[-1]['id'], wait_for=True)

    # Transaction failed if delegation period is in progress
    with pytest.raises(DryRunRevertError):
        tx_res = skale.delegation_controller.request_undelegation(
            delegation_id, wait_for=True, raise_for_status=False
        )
        tx_res.raise_for_status()

    # Skip time longer than delegation period
    _skip_evm_time(skale.web3, MONTH_IN_SECONDS * (D_DELEGATION_PERIOD + 1))

    skale.delegation_controller.request_undelegation(
        delegation_id, wait_for=True, raise_for_status=False
    )

    delegations = skale.delegation_controller.get_all_delegations_by_validator(
        validator_id=validator_id
    )
    assert delegations[-1]['id'] == delegation_id
    assert delegations[-1]['status'] == DelegationStatus.UNDELEGATION_REQUESTED



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_utils.address import to_checksum_address

from skale.fair_config.utils import convert_to_node_for_chain_config
from skale.fair_manager import FairManager
from skale.types.committee import Committee, CommitteeGroup, CommitteeIndex, Timestamp
from skale.types.node import FairNodeForChainConfig, NodeId, get_ghost_fair_node
from skale.utils.constants import ZERO_ADDRESS

""" This functions are used to generate fair config 'nodes' section data"""


def get_committee_nodes(fair: FairManager, committee_index: int) -> list[FairNodeForChainConfig]:
    committee_nodes = []
    for raw_id in fair.committee.get_committee(CommitteeIndex(committee_index)).node_ids:
        node_id: NodeId = NodeId(raw_id)
        if fair.nodes.active_node_exists(node_id):
            fair_node = fair.nodes.get(node_id)
        else:
            fair_node = get_ghost_fair_node(node_id)
        fair_node_for_config = convert_to_node_for_chain_config(fair, fair_node)
        committee_nodes.append(fair_node_for_config)
    return committee_nodes


def create_committee_group(
    fair: FairManager, committee_index: CommitteeIndex, committee: Committee, timestamp: Timestamp
) -> CommitteeGroup:
    staking_contract_address = to_checksum_address(ZERO_ADDRESS)
    if committee_index > 0:
        staking_contract_address = to_checksum_address(fair.staking.contract.address)
    return {
        'index': committee_index,
        'ts': timestamp,  # todod: remove, use from committee structure
        'staking_contract_address': staking_contract_address,
        'group': get_committee_nodes(fair, committee_index),
        'committee': committee,
    }


def get_nodes_from_two_operational_committees(fair: FairManager) -> list[CommitteeGroup]:
    """
    Compose a dictionary with nodes from the two operational committees.
    If it is the first committee, it will be saved both
    as first and second committee with first timestamp equal to 0
    If there are committee with timestamp in the future two latest committee will be saved.
    Otherwise latest committee will be saved twice.
    """
    latest_committee_index = fair.committee.last_committee_index()
    latest_committee = fair.committee.get_committee(CommitteeIndex(latest_committee_index))

    if latest_committee_index == 0:
        first_index = CommitteeIndex(0)
        first_committee = fair.committee.get_committee(CommitteeIndex(0))
        first_timestamp = Timestamp(0)
    else:
        latest_ts = fair.web3.eth.get_block('latest').get('timestamp', 0)
        first_index = latest_committee_index
        first_committee = latest_committee
        first_timestamp = Timestamp(0)

        if latest_ts < latest_committee.starting_timestamp:
            previous_committee_index = CommitteeIndex(latest_committee_index - 1)
            previous_committee = fair.committee.get_committee(
                CommitteeIndex(previous_committee_index)
            )
            first_index = previous_committee_index
            first_committee = previous_committee
            first_timestamp = first_committee.starting_timestamp

    second_index = latest_committee_index
    second_committee = latest_committee
    second_timestamp = second_committee.starting_timestamp

    return [
        create_committee_group(fair, first_index, first_committee, first_timestamp),
        create_committee_group(fair, second_index, second_committee, second_timestamp),
    ]



from unittest import mock

import pytest
from hexbytes import HexBytes

from skale.utils.web3_utils import (
    init_web3,
    private_key_to_address,
    private_key_to_public,
    to_checksum_address,
)
from skale.wallets.sgx_wallet import (
    MessageNotSignedError,
    SgxWallet,
    TransactionNotSentError,
    TransactionNotSignedError,
)
from tests.constants import ENDPOINT, ETH_PRIVATE_KEY, TEST_SGX_ENDPOINT
from tests.wallets.utils import BadSgxClient, SgxClient

ADDRESS = to_checksum_address(private_key_to_address(ETH_PRIVATE_KEY))


@pytest.fixture
@mock.patch('skale.wallets.sgx_wallet.SgxClient', new=SgxClient)
def wallet(web3):
    return SgxWallet(TEST_SGX_ENDPOINT, web3)


def test_sgx_sign(wallet):
    tx_dict = {
        'to': '0x1057dc7277a319927D3eB43e05680B75a00eb5f4',
        'value': 9,
        'gas': 200000,
        'gasPrice': 1,
        'nonce': 7,
        'chainId': None,
        'data': b'\x9b\xd9\xbb\xc6\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x95qY\xc4i\xfc;\xba\xa8\xe3\x9e\xe0\xa3$\xc28\x8a\xd6Q\xe5\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\r\xe0\xb6\xb3\xa7d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x006\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa8\xc0\x04/Rglamorous-kitalpha\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',  # noqa
    }
    wallet.sign(tx_dict)


def test_sgx_sign_and_send_without_nonce(wallet):
    send_tx_mock = mock.Mock()
    send_tx_mock.return_value = HexBytes('')
    wallet._web3.eth.send_raw_transaction = send_tx_mock
    wallet._web3.eth.get_transaction_count = mock.Mock(return_value=0)
    tx_dict = {
        'to': '0x1057dc7277a31',
        'value': 9,
        'gas': 200000,
        'gasPrice': 1,
        'chainId': None,
        'data': b'\x9b\xd9\xbb\xc6\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x95qY\xc4i\xfc;\xba\xa8\xe3\x9e\xe0\xa3$\xc28\x8a\xd6Q\xe5\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\r\xe0\xb6\xb3\xa7d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x006\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa8\xc0\x04/Rglamorous-kitalpha\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',  # noqa
    }
    signed = wallet.sign(tx_dict)
    wallet.sign_and_send(tx_dict)
    send_tx_mock.assert_called_with(signed.raw_transaction)


@mock.patch('skale.wallets.sgx_wallet.SgxClient', new=SgxClient)
def test_sgx_sign_with_key(web3):
    wallet = SgxWallet(TEST_SGX_ENDPOINT, web3, key_name='TEST_KEY')
    tx_dict = {
        'to': '0x1057dc7277a319',
        'value': 10,
        'gas': 200000,
        'gasPrice': 1,
        'nonce': 7,
        'chainId': None,
        'data': b'\x9b\xd9\xbb\xc6\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x95qY\xc4i\xfc;\xba\xa8\xe3\x9e\xe0\xa3$\xc28\x8a\xd6Q\xe5\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\r\xe0\xb6\xb3\xa7d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x006\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa8\xc0\x04/Rglamorous-kitalpha\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',  # noqa
    }
    wallet.sign(tx_dict)


@mock.patch('skale.wallets.sgx_wallet.SgxClient', new=SgxClient)
def test_sgx_sign_hash(web3):
    wallet = SgxWallet(TEST_SGX_ENDPOINT, web3, key_name='TEST_KEY')
    unsigned_hash = '0x31323331'
    signed_message = wallet.sign_hash(unsigned_hash)
    assert signed_message.signature == HexBytes('0x6161616161613131313131')


@mock.patch('skale.wallets.sgx_wallet.SgxClient', new=SgxClient)
def test_sgx_key_init(web3):
    wallet = SgxWallet(TEST_SGX_ENDPOINT, web3, 'TEST_KEY')
    pk = private_key_to_public(ETH_PRIVATE_KEY)
    assert wallet.key_name == 'TEST_KEY'
    assert wallet.address == ADDRESS
    assert wallet.public_key == pk


@mock.patch('skale.wallets.sgx_wallet.SgxClient', new=SgxClient)
def test_not_sent_error():
    w3 = init_web3(ENDPOINT)
    wallet = SgxWallet(TEST_SGX_ENDPOINT, w3, 'TEST_KEY')
    tx_dict = {
        'to': '0x1057dc7277a319927D3eB43e05680B75a00eb5f4',
        'value': 9,
        'gas': 200000,
        'gasPrice': 1,
        'nonce': 1,  # nonce too low
        'chainId': None,
        'data': b'\x9b\xd9\xbb\xc6\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x95qY\xc4i\xfc;\xba\xa8\xe3\x9e\xe0\xa3$\xc28\x8a\xd6Q\xe5\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\r\xe0\xb6\xb3\xa7d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x006\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa8\xc0\x04/Rglamorous-kitalpha\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',  # noqa
    }

    with pytest.raises(TransactionNotSentError):
        wallet.sign_and_send(tx_dict)


@mock.patch('skale.wallets.sgx_wallet.SgxClient', new=BadSgxClient)
def test_sign_error(web3):
    wallet = SgxWallet(TEST_SGX_ENDPOINT, web3, 'TEST_KEY')
    tx_dict = {
        'to': '0x1057dc7277a319927D3eB43e05680B75a00eb5f4',
        'value': 9,
        'gas': 200000,
        'gasPrice': 1,
        'nonce': 7,
        'chainId': None,
        'data': b'\x9b\xd9\xbb\xc6\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x95qY\xc4i\xfc;\xba\xa8\xe3\x9e\xe0\xa3$\xc28\x8a\xd6Q\xe5\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\r\xe0\xb6\xb3\xa7d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x006\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa8\xc0\x04/Rglamorous-kitalpha\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',  # noqa
    }
    with pytest.raises(TransactionNotSignedError):
        wallet.sign(tx_dict)


@mock.patch('skale.wallets.sgx_wallet.SgxClient', new=BadSgxClient)
def test_sign_hash_error(web3):
    wallet = SgxWallet(TEST_SGX_ENDPOINT, web3, 'TEST_KEY')
    with pytest.raises(MessageNotSignedError):
        wallet.sign_hash('0x')



import socket
from unittest.mock import Mock

from skale.fair_config.committee_nodes import get_nodes_from_two_operational_committees
from skale.types.committee import Committee, CommitteeIndex, Timestamp
from skale.types.dkg import DkgId, Fp2Point, G2Point
from skale.types.node import FairNode, NodeId


def create_mock_committee(node_ids, dkg_id=1, starting_timestamp=1000):
    return Committee(
        node_ids=node_ids,
        dkg_id=DkgId(dkg_id),
        common_public_key=G2Point(x=Fp2Point(0, 0), y=Fp2Point(0, 0)),
        starting_timestamp=Timestamp(starting_timestamp),
    )


def create_mock_node(node_id, address='0x123', name=None):
    if name is None:
        name = f'node-{node_id}'

    mock_node = Mock(spec=FairNode)
    mock_node.id = NodeId(node_id)
    mock_node.ip = socket.inet_aton(f'127.0.0.{node_id}')
    mock_node.ip_str = f'127.0.0.{node_id}'
    mock_node.domain_name = f'domain-{node_id}'
    mock_node.address = address
    mock_node.port = 10000
    mock_node.name = name
    mock_node.to_dict.return_value = {
        'id': node_id,
        'ip_str': f'127.0.0.{node_id}',
        'domain_name': f'domain-{node_id}',
        'address': address,
        'port': 10000,
        'name': name,
    }
    return mock_node


def test_get_nodes_from_last_two_committees_first_committee():
    fair = Mock()
    fair.committee.last_committee_index.return_value = 0

    node_ids = [NodeId(1), NodeId(2)]
    committee_0 = create_mock_committee(node_ids, starting_timestamp=1500)

    def get_committee_side_effect(index):
        return committee_0

    fair.committee.get_committee.side_effect = get_committee_side_effect
    fair.web3.to_checksum_address.return_value = '0x0000000000000000000000000000000000000000'

    nodes = {}
    for node_id in node_ids:
        node = create_mock_node(node_id, f'0xaddress{node_id}')
        nodes[node_id] = node
    fair.nodes.get.side_effect = lambda nid: nodes[nid]
    fair.staking.get_reward_wallet.side_effect = lambda node_id: f'0xreward{node_id}'
    fair.node.get_public_key.side_effect = lambda node_id: f'0xpublickey{node_id}'

    result = get_nodes_from_two_operational_committees(fair)

    assert len(result) == 2
    assert result[0]['index'] == 0
    assert result[0]['ts'] == 0
    assert result[1]['index'] == 0
    assert result[1]['ts'] == 1500
    assert result[0]['staking_contract_address'] == '0x0000000000000000000000000000000000000000'
    assert result[1]['staking_contract_address'] == '0x0000000000000000000000000000000000000000'


def test_get_nodes_from_last_two_committees_multiple_committees():
    fair = Mock()
    fair.committee.last_committee_index.return_value = 2
    fair.nodes.active_node_exists.return_value = True
    fair.web3.eth.get_block = Mock(return_value={'timestamp': 1018})

    node_ids_1 = [NodeId(1)]
    node_ids_2 = [NodeId(2)]
    committee_0 = create_mock_committee(node_ids_1 + node_ids_2, starting_timestamp=0)
    committee_1 = create_mock_committee(node_ids_1, starting_timestamp=1000)
    committee_2 = create_mock_committee(node_ids_2, starting_timestamp=2000)

    def get_committee_side_effect(index):
        if index == CommitteeIndex(0):
            return committee_0
        elif index == CommitteeIndex(1):
            return committee_1
        elif index == CommitteeIndex(2):
            return committee_2

    fair.committee.get_committee.side_effect = get_committee_side_effect
    fair.web3.to_checksum_address.return_value = '0x0000000000000000000000000000000000000000'
    fair.staking.contract.address = '0x7777777777777777777777777777777777777777'

    nodes = {}
    for node_id in node_ids_1 + node_ids_2:
        nodes[node_id] = create_mock_node(node_id)
    fair.nodes.get.side_effect = lambda nid: nodes[nid]
    fair.staking.get_reward_wallet.return_value = '0xreward'
    fair.node.get_public_key.side_effect = lambda node_id: f'0xpublickey{node_id}'

    result = get_nodes_from_two_operational_committees(fair)

    assert len(result) == 2
    assert result[0]['index'] == 1
    assert result[0]['ts'] == 1000
    assert result[1]['index'] == 2
    assert result[1]['ts'] == 2000
    assert result[0]['staking_contract_address'] == '0x7777777777777777777777777777777777777777'
    assert result[1]['staking_contract_address'] == '0x7777777777777777777777777777777777777777'


def test_get_nodes_from_last_two_committees_structure():
    fair = Mock()
    fair.committee.last_committee_index.return_value = 0

    node_ids = [NodeId(1)]
    committee = create_mock_committee(node_ids)

    def get_committee_side_effect(index):
        return committee

    fair.committee.get_committee.side_effect = get_committee_side_effect
    fair.web3.to_checksum_address.return_value = '0x0000000000000000000000000000000000000000'
    fair.nodes.get.return_value = create_mock_node(1)
    fair.staking.get_reward_wallet.return_value = '0xreward'
    fair.node.get_public_key.side_effect = lambda node_id: f'0xpublickey{node_id}'

    result = get_nodes_from_two_operational_committees(fair)

    for group in result:
        assert 'index' in group
        assert 'ts' in group
        assert 'group' in group
        assert 'committee' in group
        assert isinstance(group['index'], int)
        assert isinstance(group['ts'], int)
        assert isinstance(group['group'], list)
        assert isinstance(group['committee'], Committee)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2024-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import socket
from dataclasses import dataclass
from enum import IntEnum
from typing import List, NewType, TypedDict, TypeVar

from eth_typing import BlockNumber, ChecksumAddress, HexStr
from eth_utils.address import to_checksum_address

from skale.types.schain import SchainHash
from skale.types.validator import ValidatorId
from skale.utils.constants import ZERO_ADDRESS

NodeId = NewType('NodeId', int)
Port = NewType('Port', int)


class NodeStatus(IntEnum):
    ACTIVE = 0
    LEAVING = 1
    LEFT = 2
    IN_MAINTENANCE = 3


class NodeBase(TypedDict):
    name: str
    port: Port
    start_block: BlockNumber
    last_reward_date: int
    finish_time: int
    status: NodeStatus
    validator_id: ValidatorId
    publicKey: HexStr
    domain_name: str


class Node(NodeBase):
    ip: bytes
    publicIP: bytes


FairNodeType = TypeVar('FairNodeType', bound='FairNode')


@dataclass
class FairNode:
    id: NodeId
    ip: bytes
    ip_str: str
    domain_name: str
    address: ChecksumAddress
    port: Port
    name: str

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'ip_str': self.ip_str,
            'domain_name': self.domain_name,
            'address': self.address,
            'port': self.port,
            'name': self.name,
        }


FairNodeForChainConfigType = TypeVar('FairNodeForChainConfigType', bound='FairNodeForChainConfig')


@dataclass
class FairNodeForChainConfig(FairNode):
    reward_wallet_address: ChecksumAddress
    public_key: HexStr

    def to_dict(self) -> dict:
        return {
            'reward_wallet_address': self.reward_wallet_address,
            'public_key': self.public_key,
            **super().to_dict(),
        }


def get_ghost_fair_node(node_id: NodeId) -> FairNode:
    ip_str = '255.255.255.255'
    return FairNode(
        id=node_id,
        ip=socket.inet_aton(ip_str),
        ip_str=ip_str,
        domain_name='',
        address=to_checksum_address(ZERO_ADDRESS),
        port=Port(10000),
        name='',
    )


class NodeWithId(Node):
    id: NodeId


class NodeWithChangeIp(NodeBase):
    id: NodeId
    ip_change_ts: int
    ip: str
    publicIP: str


class NodeWithSchainHashes(NodeWithId):
    schain_hashes: List[SchainHash]



"""SKALE node rotation test"""

from unittest import mock

import pytest

from skale.contracts.manager.node_rotation import Rotation
from skale.utils.contracts_provision.main import (
    add_test4_schain_type,
    cleanup_nodes_schains,
    create_schain,
)
from tests.constants import DEFAULT_SCHAIN_HASH, DEFAULT_SCHAIN_INDEX, DEFAULT_SCHAIN_NAME
from tests.rotation_history.utils import TEST_ROTATION_DELAY, _skip_evm_time, run_dkg, set_up_nodes


def test_get_rotation(skale):
    assert skale.node_rotation.get_rotation(DEFAULT_SCHAIN_NAME) == Rotation(
        leaving_node_id=0, new_node_id=0, freeze_until=0, rotation_counter=0
    )


def test_get_leaving_history(skale):
    empty = skale.node_rotation.get_leaving_history(DEFAULT_SCHAIN_INDEX)
    assert empty == []
    with mock.patch.object(
        skale.node_rotation.contract.functions.getLeavingHistory, 'call'
    ) as call_mock:
        call_mock.return_value = [(DEFAULT_SCHAIN_HASH, 1000), (DEFAULT_SCHAIN_HASH, 2000)]
        history = skale.node_rotation.get_leaving_history(DEFAULT_SCHAIN_INDEX)
        assert isinstance(history, list)
        assert history == [
            {'schain_id': DEFAULT_SCHAIN_HASH, 'finished_rotation': 1000},
            {'schain_id': DEFAULT_SCHAIN_HASH, 'finished_rotation': 2000},
        ]


def test_is_rotation_in_progress(skale):
    assert skale.node_rotation.is_rotation_in_progress(DEFAULT_SCHAIN_NAME) is False


def test_wait_for_new_node(skale):
    assert skale.node_rotation.wait_for_new_node(DEFAULT_SCHAIN_NAME) is False


@pytest.fixture
def four_node_schain(skale, validator):
    nodes, skale_instances = set_up_nodes(skale, 4, no_zero_id=False)
    add_test4_schain_type(skale)
    try:
        name = create_schain(
            skale,
            schain_type=2,  # test4 should have 2 index
            random_name=True,
        )
        yield nodes, skale_instances, name
    finally:
        cleanup_nodes_schains(skale)


def test_is_rotation_active(skale, four_node_schain):
    nodes, skale_instances, name = four_node_schain
    group_index = skale.web3.keccak(text=name)

    run_dkg(nodes, skale_instances, group_index)

    exiting_node_index = 3
    new_nodes, new_skale_instances = set_up_nodes(skale, 1, no_zero_id=False)

    assert not skale.node_rotation.is_new_node_found(name)
    assert not skale.node_rotation.is_rotation_in_progress(name)
    assert not skale.node_rotation.is_rotation_active(name)

    skale.nodes.init_exit(nodes[exiting_node_index]['node_id'])

    assert not skale.node_rotation.is_new_node_found(name)
    assert skale.node_rotation.is_rotation_in_progress(name)
    assert not skale.node_rotation.is_rotation_active(name)

    skale_instances[exiting_node_index].manager.node_exit(nodes[exiting_node_index]['node_id'])

    assert skale.node_rotation.is_new_node_found(name)
    assert skale.node_rotation.is_rotation_in_progress(name)
    assert skale.node_rotation.is_rotation_active(name)

    nodes[exiting_node_index] = new_nodes[0]
    skale_instances[exiting_node_index] = new_skale_instances[0]

    run_dkg(nodes, skale_instances, group_index, skip_time=False, rotation_id=1)

    assert skale.node_rotation.is_new_node_found(name)
    assert skale.node_rotation.is_rotation_in_progress(name)
    assert skale.node_rotation.is_rotation_active(name)

    _skip_evm_time(skale.web3, TEST_ROTATION_DELAY)

    assert skale.node_rotation.is_new_node_found(name)
    assert not skale.node_rotation.is_rotation_in_progress(name)
    assert not skale.node_rotation.is_rotation_active(name)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging
from functools import wraps
from typing import Any, Callable, cast

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.types import Nonce, Wei

import skale.config as config
from skale.transactions.result import TxRes, TxStatus
from skale.transactions.tools import make_dry_run_call, transaction_from_method
from skale.utils.helper import to_camel_case
from skale.utils.web3_utils import (
    DEFAULT_BLOCKS_TO_WAIT,
    MAX_WAITING_TIME,
    default_gas_price,
    get_eth_nonce,
    wait_for_confirmation_blocks,
)
from skale.wallets.common import BaseWallet

logger = logging.getLogger(__name__)


class BaseContract:
    def __init__(
        self,
        web3: Web3,
        address: ChecksumAddress,
        abi: list[Any],
        wallet: BaseWallet | None = None,
    ):
        self.web3 = web3
        self._wallet = wallet
        self._init_contract(address, abi)

    @property
    def wallet(self) -> BaseWallet | None:
        return self._wallet

    @wallet.setter
    def wallet(self, value: BaseWallet | None) -> None:
        self._wallet = value

    def _init_contract(self, address: ChecksumAddress, abi: list[Any]) -> None:
        self.address = Web3.to_checksum_address(address)
        self.contract = self.web3.eth.contract(address=self.address, abi=abi)

    def __getattr__(self, attr: str) -> Callable[..., Any]:
        """Fallback for contract calls"""
        logger.debug('Calling contract function: %s', attr)

        def wrapper(*args: Any, **kw: Any) -> Any:
            logger.debug('called with %r and %r' % (args, kw))
            camel_case_fn_name = to_camel_case(attr)
            if hasattr(self.contract.functions, camel_case_fn_name):
                return getattr(self.contract.functions, camel_case_fn_name)(*args, **kw).call()
            if hasattr(self.contract.functions, attr):
                return getattr(self.contract.functions, attr)(*args, **kw).call()
            raise AttributeError(attr)

        return wrapper


def transaction_method(transaction: Callable[..., ContractFunction]) -> Callable[..., TxRes]:
    @wraps(transaction)
    def wrapper(
        self: BaseContract,
        *args: Any,
        wait_for: bool = True,
        blocks_to_wait: int = DEFAULT_BLOCKS_TO_WAIT,
        timeout: int = MAX_WAITING_TIME,
        gas_limit: int | None = None,
        gas_price: int | None = None,
        nonce: Nonce | None = None,
        max_fee_per_gas: int | None = None,
        max_priority_fee_per_gas: int | None = None,
        value: Wei = Wei(0),
        dry_run_only: bool = False,
        skip_dry_run: bool = False,
        raise_for_status: bool = True,
        multiplier: float | None = None,
        priority: int | None = None,
        confirmation_blocks: int = 0,
        **kwargs: Any,
    ) -> TxRes:
        if not self.wallet:
            raise ValueError('Wallet is not set for this contract')
        method = transaction(self, *args, **kwargs)

        nonce = get_eth_nonce(self.web3, self.wallet.address)

        call_result, tx_hash, receipt = None, None, None
        should_dry_run = not skip_dry_run and not config.DISABLE_DRY_RUN

        dry_run_success = False
        if should_dry_run:
            call_result = make_dry_run_call(self.web3, self.wallet, method, gas_limit, value)
            if call_result.status == TxStatus.SUCCESS:
                gas_limit = gas_limit or int(call_result.data['gas'])
                dry_run_success = True

        should_send = not dry_run_only and (not should_dry_run or dry_run_success)
        if should_send:
            gas_limit = gas_limit or config.DEFAULT_GAS_LIMIT
            gas_price = gas_price or config.DEFAULT_GAS_PRICE_WEI or default_gas_price(self.web3)
            tx = transaction_from_method(
                method=method,
                gas_limit=gas_limit,
                gas_price=gas_price,
                max_fee_per_gas=max_fee_per_gas,
                max_priority_fee_per_gas=max_priority_fee_per_gas,
                nonce=nonce,
                value=value,
            )
            method_name = f'{self.name}.{method.abi.get("name")}'
            tx_hash = self.wallet.sign_and_send(
                tx, multiplier=multiplier, priority=priority, method=method_name
            )

        if tx_hash is not None and wait_for:
            receipt = self.wallet.wait(tx_hash)

        should_confirm = receipt is not None and confirmation_blocks > 0
        if should_confirm:
            wait_for_confirmation_blocks(self.web3, confirmation_blocks)

        tx_res = TxRes(call_result, tx_hash, receipt)

        if raise_for_status:
            tx_res.raise_for_status()
        return tx_res

    # return wrapper
    return cast(Callable[..., TxRes], wrapper)



"""Tests for skale/allocator/allocator.py"""

import pytest

from skale.utils.account_tools import send_eth
from skale.utils.contracts_provision import D_PLAN_ID
from skale.utils.contracts_provision.allocator import add_test_plan, connect_test_beneficiary
from skale.wallets.web3_wallet import generate_wallet

TEST_PLAN_PARAMS = {
    'totalVestingDuration': 36,
    'vestingCliff': 6,
    'vestingIntervalTimeUnit': 1,
    'vestingInterval': 6,
    'isDelegationAllowed': True,
    'isTerminatable': True,
}

D_END_TIMESTAMP = 1693526400
D_END_LOCKUP_TIMESTAMP = 1614556800


def test_is_beneficiary_registered(skale_allocator):
    wallet = generate_wallet(skale_allocator.web3)
    assert not skale_allocator.allocator.is_beneficiary_registered(wallet.address)


def test_add_plan(skale_allocator):
    plan_id = add_test_plan(skale_allocator)
    assert isinstance(plan_id, int)


def test_connect_beneficiary_to_plan(skale_allocator):
    wallet = generate_wallet(skale_allocator.web3)
    assert not skale_allocator.allocator.is_beneficiary_registered(wallet.address)
    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)
    assert skale_allocator.allocator.is_beneficiary_registered(wallet.address)


def test_start_vesting(skale_allocator):
    main_wallet = skale_allocator.wallet
    wallet = generate_wallet(skale_allocator.web3)

    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)
    send_eth(skale_allocator.web3, main_wallet, wallet.address, 0.1)

    assert skale_allocator.allocator.is_delegation_allowed(wallet.address)

    assert not skale_allocator.allocator.is_vesting_active(wallet.address)
    skale_allocator.allocator.start_vesting(wallet.address, wait_for=True)
    assert skale_allocator.allocator.is_vesting_active(wallet.address)


def test_stop_vesting(skale_allocator):
    # todo: impove test
    main_wallet = skale_allocator.wallet
    wallet = generate_wallet(skale_allocator.web3)

    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)
    send_eth(skale_allocator.web3, main_wallet, wallet.address, 0.1)

    assert skale_allocator.allocator.is_delegation_allowed(wallet.address)

    assert not skale_allocator.allocator.is_vesting_active(wallet.address)
    skale_allocator.allocator.start_vesting(wallet.address, wait_for=True)
    assert skale_allocator.allocator.is_vesting_active(wallet.address)

    skale_allocator.allocator.stop_vesting(wallet.address, wait_for=True)
    assert not skale_allocator.allocator.is_vesting_active(wallet.address)


def test_grant_vesting_manager_role(skale_allocator):
    wallet = generate_wallet(skale_allocator.web3)
    vesting_manager_role = skale_allocator.allocator.vesting_manager_role()
    assert not skale_allocator.allocator.has_role(vesting_manager_role, wallet.address)
    skale_allocator.allocator.grant_role(vesting_manager_role, wallet.address)
    assert skale_allocator.allocator.has_role(vesting_manager_role, wallet.address)


def test_get_beneficiary_plan_params(skale_allocator):
    wallet = generate_wallet(skale_allocator.web3)
    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)
    beneficiary = skale_allocator.allocator.get_beneficiary_plan_params(wallet.address)
    assert beneficiary['planId'] == 1
    assert beneficiary['startMonth'] == 8
    assert beneficiary['amountAfterLockup'] == 1000000000000000000000


def test_get_plan(skale_allocator):
    plan = skale_allocator.allocator.get_plan(D_PLAN_ID)
    assert plan == TEST_PLAN_PARAMS


def test_get_all_plans(skale_allocator):
    plans = skale_allocator.allocator.get_all_plans()
    plan_params_with_id = TEST_PLAN_PARAMS.copy()
    plan_params_with_id['planId'] = 1
    assert plan_params_with_id == plans[0]


@pytest.mark.skip('test should be updated')
def test_calculate_vested_amount(skale_allocator):
    wallet = generate_wallet(skale_allocator.web3)
    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)
    vested_amount = skale_allocator.allocator.calculate_vested_amount(wallet.address)
    assert vested_amount == 1000000000000000000000


def test_get_finish_vesting_time(skale_allocator):
    wallet = generate_wallet(skale_allocator.web3)
    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)
    res = skale_allocator.allocator.get_finish_vesting_time(wallet.address)
    assert res == D_END_TIMESTAMP


def test_get_lockup_period_end_timestamp(skale_allocator):
    wallet = generate_wallet(skale_allocator.web3)
    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)
    res = skale_allocator.allocator.get_lockup_period_end_timestamp(wallet.address)
    assert res == D_END_LOCKUP_TIMESTAMP


@pytest.mark.skip('test should be updated')
def test_get_time_of_next_vest(skale_allocator):
    wallet = generate_wallet(skale_allocator.web3)
    connect_test_beneficiary(skale_allocator, D_PLAN_ID, wallet)
    res = skale_allocator.allocator.get_time_of_next_vest(wallet.address)
    assert res == 1630454400



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import cast

from eth_account import messages
from eth_account.datastructures import SignedMessage, SignedTransaction
from eth_keys.datatypes import PublicKey
from eth_keys.main import lazy_key_api as keys
from eth_typing import AnyAddress, ChecksumAddress, HexStr
from web3 import Web3
from web3.exceptions import Web3Exception
from web3.types import TxParams, TxReceipt, _Hash32

import skale.config as config
from skale.transactions.exceptions import TransactionNotSentError, TransactionNotSignedError
from skale.utils.web3_utils import (
    DEFAULT_BLOCKS_TO_WAIT,
    MAX_WAITING_TIME,
    get_eth_nonce,
    wait_for_receipt_by_blocks,
)
from skale.wallets.common import BaseWallet, MessageNotSignedError, ensure_chain_id


def private_key_to_public(pr: HexStr) -> PublicKey:
    pr_bytes = Web3.to_bytes(hexstr=pr)
    pk = keys.PrivateKey(pr_bytes)
    return pk.public_key


def public_key_to_address(pk: PublicKey) -> ChecksumAddress:
    hash = Web3.keccak(hexstr=str(pk))
    return to_checksum_address(Web3.to_hex(hash[-20:]))


def private_key_to_address(pr: HexStr) -> ChecksumAddress:
    pk = private_key_to_public(pr)
    return public_key_to_address(pk)


def to_checksum_address(address: AnyAddress | str | bytes) -> ChecksumAddress:
    return Web3.to_checksum_address(address)


class Web3Wallet(BaseWallet):
    def __init__(self, private_key: HexStr, web3: Web3):
        self._private_key = private_key
        self._public_key = private_key_to_public(self._private_key)
        self._address = public_key_to_address(self._public_key)

        self._web3 = web3

    def sign(self, tx_dict: TxParams) -> SignedTransaction:
        if not tx_dict.get('nonce'):
            tx_dict['nonce'] = get_eth_nonce(self._web3, self._address)
        ensure_chain_id(tx_dict, self._web3)
        try:
            return cast(
                SignedTransaction,
                self._web3.eth.account.sign_transaction(tx_dict, private_key=self._private_key),
            )
        except (TypeError, ValueError, Web3Exception) as e:
            raise TransactionNotSignedError(e)

    def sign_hash(self, unsigned_hash: str) -> SignedMessage:
        try:
            unsigned_message = messages.encode_defunct(hexstr=unsigned_hash)
            return cast(
                SignedMessage,
                self._web3.eth.account.sign_message(
                    unsigned_message, private_key=self._private_key
                ),
            )
        except (TypeError, ValueError, Web3Exception) as e:
            raise MessageNotSignedError(e)

    def sign_and_send(
        self,
        tx_dict: TxParams,
        multiplier: float | None = config.DEFAULT_GAS_MULTIPLIER,
        priority: int | None = config.DEFAULT_PRIORITY,
        method: str | None = None,
    ) -> HexStr:
        signed_tx = self.sign(tx_dict)
        try:
            return Web3.to_hex(self._web3.eth.send_raw_transaction(signed_tx.raw_transaction))
        except (ValueError, Web3Exception) as e:
            raise TransactionNotSentError(e)

    @property
    def address(self) -> ChecksumAddress:
        return self._address

    @property
    def public_key(self) -> str:
        return str(self._public_key)

    def wait(
        self,
        tx_hash: _Hash32,
        blocks_to_wait: int = DEFAULT_BLOCKS_TO_WAIT,
        timeout: int = MAX_WAITING_TIME,
    ) -> TxReceipt:
        return wait_for_receipt_by_blocks(
            self._web3, tx_hash, blocks_to_wait=blocks_to_wait, timeout=timeout
        )


def generate_wallet(web3: Web3) -> Web3Wallet:
    account = web3.eth.account.create()
    private_key = account.key.hex()
    return Web3Wallet(private_key, web3)


def generate_wallets(web3: Web3, n_of_keys: int) -> list[Web3Wallet]:
    return [generate_wallet(web3) for _ in range(n_of_keys)]



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
"""SKALE Contract manager class"""

from Crypto.Hash import keccak
from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_manager_contract import SkaleManagerContract
from skale.utils.helper import add_0x_prefix


class ContractManager(SkaleManagerContract):
    @transaction_method
    def set_contracts_address(
        self, contracts_name: str, contracts_address: ChecksumAddress
    ) -> 'ContractFunction':
        return self.contract.functions.setContractsAddress(contracts_name, contracts_address)

    def get_contract_address(self, name: str) -> ChecksumAddress:
        contract_hash = add_0x_prefix(self.get_contract_hash_by_name(name))
        return Web3.to_checksum_address(self.contract.functions.contracts(contract_hash).call())

    def get_contract_hash_by_name(self, name: str) -> str:
        keccak_hash = keccak.new(data=name.encode('utf8'), digest_bits=256)
        return keccak_hash.hexdigest()



import pytest
from web3.exceptions import ContractLogicError

from skale.transactions.exceptions import DryRunRevertError

START_IP = '0.0.0.1'
END_IP = '255.255.255.254'


@pytest.fixture
def sync_manager_permissions(skale):
    skale.sync_manager.grant_sync_manager_role(skale.wallet.address)


@pytest.fixture
def ip_range(skale):
    name = 'test-range'
    skale.sync_manager.add_ip_range(name, START_IP, END_IP)
    try:
        yield name
    finally:
        skale.sync_manager.remove_ip_range(name)


def test_add_get_remove_ip_range(skale, sync_manager_permissions, ip_range, block_in_seconds):
    r = skale.sync_manager.get_ip_range_by_name(ip_range)
    assert r.start_ip == START_IP
    assert r.end_ip == END_IP

    r = skale.sync_manager.get_ip_range_by_index(0)
    assert r.start_ip == START_IP
    assert r.end_ip == END_IP


def test_add_bad_ip_range(skale, sync_manager_permissions, block_in_seconds):
    with pytest.raises(DryRunRevertError):
        skale.sync_manager.add_ip_range('brange', '0.0.0.0', '1.1.1.1')

    with pytest.raises(DryRunRevertError):
        skale.sync_manager.add_ip_range('brange', '2.2.2.2', '1.1.1.1')

    with pytest.raises(OSError):
        skale.sync_manager.add_ip_range('brange', '1.1.1.1', '256.256.256.256')


def test_remove_range_bad_params(skale, sync_manager_permissions, block_in_seconds):
    with pytest.raises(DryRunRevertError):
        skale.sync_manager.remove_ip_range('phantom')


def test_get_range_bad_params(skale, sync_manager_permissions, block_in_seconds):
    num = skale.sync_manager.get_ip_ranges_number()
    # TODO: Make dry run handle revert that has empty reason properly
    with pytest.raises(ContractLogicError):
        r = skale.sync_manager.get_ip_range_by_index(num)
    with pytest.raises(ContractLogicError):
        r = skale.sync_manager.get_ip_range_by_index(0)
    r = skale.sync_manager.get_ip_range_by_name('phantom')
    assert r.start_ip == '0.0.0.0' and r.end_ip == '0.0.0.0'



import importlib
import os
from unittest import mock

import pytest
from web3 import Web3

import skale.config as config
from skale.transactions.exceptions import TransactionNotSentError
from skale.transactions.result import TxStatus
from skale.transactions.tools import estimate_gas
from skale.utils.account_tools import generate_account
from skale.utils.contracts_provision.utils import generate_random_schain_data
from skale.utils.web3_utils import wait_for_receipt_by_blocks
from tests.constants import TEST_GAS_LIMIT

ETH_IN_WEI = 10**18
CUSTOM_DEFAULT_GAS_LIMIT = 2 * 10**6
CUSTOM_DEFAULT_GAS_PRICE_WEI = 1500000000


def test_dry_run(skale):
    account = generate_account(skale.web3)
    address_to = account['address']
    address_from = Web3.to_checksum_address(skale.wallet.address)
    address_to = Web3.to_checksum_address(address_to)
    balance_from_before = skale.token.get_balance(address_from)
    balance_to_before = skale.token.get_balance(address_to)
    amount = 10 * ETH_IN_WEI
    tx_res = skale.token.transfer(address_to, amount, dry_run_only=True)
    assert isinstance(tx_res.tx_call_result.data['gas'], int)
    assert tx_res.tx_call_result.status == TxStatus.SUCCESS
    tx_res.raise_for_status()

    balance_from_after = skale.token.get_balance(address_from)
    assert balance_from_after == balance_from_before
    balance_to_after = skale.token.get_balance(address_to)
    assert balance_to_after == balance_to_before


@pytest.fixture
def disable_dry_run_env():
    os.environ['DISABLE_DRY_RUN'] = 'True'
    os.environ['DEFAULT_GAS_LIMIT'] = str(CUSTOM_DEFAULT_GAS_LIMIT)
    os.environ['DEFAULT_GAS_PRICE_WEI'] = str(CUSTOM_DEFAULT_GAS_PRICE_WEI)
    importlib.reload(config)
    yield
    os.environ.pop('DISABLE_DRY_RUN')
    os.environ.pop('DEFAULT_GAS_LIMIT')
    os.environ.pop('DEFAULT_GAS_PRICE_WEI')
    importlib.reload(config)


def test_disable_dry_run_env(skale, disable_dry_run_env):
    account = generate_account(skale.web3)
    address_to = account['address']
    amount = 10 * ETH_IN_WEI
    with mock.patch('skale.contracts.base_contract.make_dry_run_call') as dry_run_mock:
        skale.token.transfer(address_to, amount)
        dry_run_mock.assert_not_called()


def test_skip_dry_run(skale):
    account = generate_account(skale.web3)
    address_to = account['address']
    address_from = Web3.to_checksum_address(skale.wallet.address)
    address_to = Web3.to_checksum_address(address_to)
    balance_from_before = skale.token.get_balance(address_from)
    balance_to_before = skale.token.get_balance(address_to)
    amount = 10 * ETH_IN_WEI

    tx_res = skale.token.transfer(address_to, amount, skip_dry_run=True, gas_limit=TEST_GAS_LIMIT)
    assert tx_res.tx_hash is not None, tx_res
    assert tx_res.receipt is not None
    assert tx_res.tx_call_result is None
    balance_from_after = skale.token.get_balance(address_from)
    assert balance_from_after == balance_from_before - amount
    balance_to_after = skale.token.get_balance(address_to)
    assert balance_to_after == balance_to_before + amount


def test_wait_for_false(skale):
    ETH_IN_WEI = 10**18
    account = generate_account(skale.web3)
    address_to = account['address']
    address_from = Web3.to_checksum_address(skale.wallet.address)
    address_to = Web3.to_checksum_address(address_to)
    balance_from_before = skale.token.get_balance(address_from)
    balance_to_before = skale.token.get_balance(address_to)
    amount = 10 * ETH_IN_WEI

    tx_res = skale.token.transfer(address_to, amount, wait_for=False)
    assert tx_res.tx_hash is not None
    assert tx_res.receipt is None
    assert isinstance(tx_res.tx_call_result.data['gas'], int)
    assert tx_res.tx_call_result.status == TxStatus.SUCCESS

    tx_res.receipt = wait_for_receipt_by_blocks(skale.web3, tx_res.tx_hash)
    tx_res.raise_for_status()

    balance_from_after = skale.token.get_balance(address_from)
    assert balance_from_after == balance_from_before - amount
    balance_to_after = skale.token.get_balance(address_to)
    assert balance_to_after == balance_to_before + amount


def test_tx_res_dry_run(skale):
    account = generate_account(skale.web3)
    token_amount = 10
    tx_res = skale.token.transfer(account['address'], token_amount, dry_run_only=True)
    assert tx_res.tx_call_result is not None
    assert tx_res.tx_hash is None
    assert tx_res.receipt is None
    tx_res.raise_for_status()


def test_tx_res_wait_for_false(skale):
    account = generate_account(skale.web3)
    token_amount = 10
    tx_res = skale.token.transfer(account['address'], token_amount, wait_for=False)
    assert tx_res.tx_hash is not None
    assert tx_res.receipt is None
    tx_res.raise_for_status()

    tx_res.receipt = wait_for_receipt_by_blocks(skale.web3, tx_res.tx_hash)
    tx_res.raise_for_status()


def test_tx_res_wait_for_true(skale):
    account = generate_account(skale.web3)
    token_amount = 10
    tx_res = skale.token.transfer(account['address'], token_amount)
    assert tx_res.tx_hash is not None
    assert tx_res.receipt is not None
    tx_res.raise_for_status()


def test_tx_res_with_insufficient_funds(skale):
    account = generate_account(skale.web3)
    token_amount = 9
    huge_gas_price = 10**22
    with pytest.raises(TransactionNotSentError):
        skale.token.transfer(account['address'], token_amount, gas_price=huge_gas_price)


def test_confirmation_blocks(skale):
    account = generate_account(skale.web3)
    token_amount = 10
    confirmation_blocks = 0  # todo: enable mining on ganache
    start_block = skale.web3.eth.block_number
    skale.token.transfer(account['address'], token_amount, confirmation_blocks=confirmation_blocks)
    assert skale.web3.eth.block_number >= start_block + confirmation_blocks


def test_block_limit_estimate_gas(skale):
    account = generate_account(skale.web3)
    token_amount = 10
    max_gas = 200000000
    with mock.patch.object(
        skale.token.contract.functions.transfer, 'estimate_gas', new=mock.Mock(return_value=max_gas)
    ):
        method = skale.token.contract.functions.transfer(account['address'], token_amount)
        res = estimate_gas(skale.web3, method, {'from': skale.wallet.address})
        assert res < max_gas


def test_value_option(skale, nodes):
    skale.schains.grant_role(skale.schains.schain_creator_role(), skale.wallet.address)
    type_of_nodes, lifetime_seconds, name = generate_random_schain_data(skale)
    value_wei = 1000
    try:
        skale.schains.add_schain_by_foundation(
            lifetime_seconds, type_of_nodes, 0, name, wait_for=True, value=value_wei
        )
    finally:
        skale.manager.delete_schain(name, wait_for=True)



"""Tests for contracts/delegation/delegation_period_manager.py"""

import random

STAKE_MULTIPLIER = 100
CUSTOM_DELEGATION_PERIOD = 51


def test_set_delegation_period(skale):
    new_delegation_period = random.randint(13, 50)
    assert not skale.delegation_period_manager.is_delegation_period_allowed(new_delegation_period)
    skale.delegation_period_manager.set_delegation_period(
        months_count=new_delegation_period, stake_multiplier=STAKE_MULTIPLIER, wait_for=True
    )
    assert skale.delegation_period_manager.is_delegation_period_allowed(new_delegation_period)


def test_is_delegation_period_allowed(skale):
    assert not skale.delegation_period_manager.is_delegation_period_allowed(
        months_count=CUSTOM_DELEGATION_PERIOD
    )



"""SKALE node test"""

import random
import socket
import string
from datetime import datetime

import pytest
from eth_keys import keys
from web3 import Web3

import skale.utils.helper as Helper
from skale.contracts.manager.nodes import FIELDS, NodeStatus
from skale.skale_manager import SkaleManager
from skale.utils.contracts_provision import DEFAULT_DOMAIN_NAME
from skale.utils.contracts_provision.utils import generate_random_ip
from skale.utils.exceptions import InvalidNodeIdError
from tests.constants import DEFAULT_NODE_ID, DEFAULT_NODE_NAME, NOT_EXISTING_ID


def test_get_raw_not_exist(skale):
    with pytest.raises(InvalidNodeIdError):
        skale.nodes._Nodes__get_raw(NOT_EXISTING_ID)


def public_key_from_private(key):
    pr_bytes = Web3.to_bytes(hexstr=key)
    return keys.PrivateKey(pr_bytes)


def test_get(skale: SkaleManager, nodes, node_wallets):
    node = skale.nodes.get_by_name(DEFAULT_NODE_NAME)
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    node_by_id = skale.nodes.get(node_id)
    assert list(node.keys()) == FIELDS
    assert [k for k, v in node.items() if v is None] == []
    assert node_by_id['name'] == DEFAULT_NODE_NAME
    socket.inet_ntoa(node_by_id['ip'])
    socket.inet_ntoa(node_by_id['publicIP'])

    assert node_by_id['publicKey'] == node_wallets[0].public_key

    assert node_by_id['publicKey'] != b''
    assert node_by_id['start_block'] > 0
    assert node_by_id['last_reward_date'] > 0
    assert node_by_id['finish_time'] == 0
    assert node_by_id['status'] == 0
    assert node_by_id['validator_id'] == 1

    assert list(node_by_id.keys()) == FIELDS
    assert node == node_by_id

    with pytest.raises(InvalidNodeIdError):
        skale.nodes.get(NOT_EXISTING_ID)


def test_get_with_change_ip(skale: SkaleManager, nodes):
    node = skale.nodes.get_with_change_ip(DEFAULT_NODE_ID)
    assert node['ip_change_ts'] is not None
    assert isinstance(node['ip_change_ts'], int)


def test_wrong_node_id(skale):
    with pytest.raises(InvalidNodeIdError):
        skale.nodes.node_status(NOT_EXISTING_ID)

    with pytest.raises(InvalidNodeIdError):
        skale.nodes.node_finish_time(NOT_EXISTING_ID)


def test_active_node_ids(skale: SkaleManager, nodes):
    nodes_number = skale.nodes.nodes_number()
    active_node_ids = skale.nodes.active_node_ids()

    assert len(active_node_ids) <= nodes_number

    assert isinstance(active_node_ids, list)
    assert len(active_node_ids) > 0

    node = skale.nodes.get(active_node_ids[-1])
    assert list(node.keys()) == FIELDS

    # 0 is active status
    assert all([skale.nodes.node_status(node_id) == 0 for node_id in active_node_ids])


def test_active_node_ips(skale: SkaleManager, nodes):
    nodes_number = skale.nodes.nodes_number()

    active_node_ips = skale.nodes.active_node_ips()
    assert len(active_node_ips) <= nodes_number

    assert isinstance(active_node_ips, list)

    assert all([Helper.is_valid_ipv4_address(node_ip) for node_ip in active_node_ips])


def test_is_node_name_available(skale: SkaleManager, nodes):
    node = skale.nodes.get_by_name(DEFAULT_NODE_NAME)
    unused_name = 'unused_name'
    assert skale.nodes.is_node_name_available(node['name']) is False
    assert skale.nodes.is_node_name_available(unused_name) is True


def test_is_node_ip_available(skale: SkaleManager, nodes):
    node = skale.nodes.get_by_name(DEFAULT_NODE_NAME)
    node_ip = Helper.ip_from_bytes(node['ip'])

    unused_ip = '123.123.231.123'
    assert skale.nodes.is_node_ip_available(node_ip) is False
    assert skale.nodes.is_node_name_available(unused_ip) is True


def test_node_name_to_index(skale: SkaleManager, nodes):
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    node_by_id_data = skale.nodes.get(node_id)
    node_by_name_data = skale.nodes.get_by_name(DEFAULT_NODE_NAME)
    assert node_by_id_data == node_by_name_data


def test_node_public_key(skale: SkaleManager, nodes, node_wallets):
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    node_public_key = skale.nodes.node_public_key(node_id)
    assert node_public_key == node_wallets[0].public_key

    with pytest.raises(InvalidNodeIdError):
        skale.nodes.node_public_key(NOT_EXISTING_ID)


def test_node_in_maintenance(skale: SkaleManager, nodes):
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    assert skale.nodes.node_status(node_id) == NodeStatus.ACTIVE.value

    try:
        skale.nodes.set_node_in_maintenance(node_id)
        assert skale.nodes.node_status(node_id) == NodeStatus.IN_MAINTENANCE.value
    finally:
        skale.nodes.remove_node_from_in_maintenance(node_id)
    assert skale.nodes.node_status(node_id) == NodeStatus.ACTIVE.value


def test_set_domain_name(skale: SkaleManager, nodes):
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    node = skale.nodes.get(node_id)
    random_domain = ''.join(random.choices(string.ascii_uppercase + string.digits, k=8))

    assert node['domain_name'] != random_domain

    skale.nodes.set_domain_name(node_id, random_domain)

    node = skale.nodes.get(node_id)
    assert node['domain_name'] == random_domain

    skale.nodes.set_domain_name(node_id, DEFAULT_DOMAIN_NAME)


def test_domain_name(skale: SkaleManager, nodes):
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    node = skale.nodes.get(node_id)
    assert node['domain_name'] == DEFAULT_DOMAIN_NAME


def test_get_node_next_reward_date(skale: SkaleManager, nodes):
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    next_reward_date_ts = skale.nodes.get_node_next_reward_date(node_id)
    next_reward_date = datetime.utcfromtimestamp(next_reward_date_ts)
    present = datetime.now()
    assert next_reward_date > present


def test_change_ip(skale: SkaleManager, nodes):
    node_id, _ = nodes
    old_ip = skale.nodes.get(node_id)['ip']

    new_ip = Helper.ip_to_bytes(generate_random_ip())

    skale.nodes.change_ip(node_id, new_ip, new_ip)
    data = skale.nodes.get(node_id)
    assert data['ip'] == new_ip
    assert data['publicIP'] == new_ip

    skale.nodes.change_ip(node_id, old_ip, old_ip)
    data = skale.nodes.get(node_id)
    assert data['ip'] == old_ip
    assert data['publicIP'] == old_ip


def test_last_change_ip_time(skale: SkaleManager, nodes):
    node_id = skale.nodes.node_name_to_index(DEFAULT_NODE_NAME)
    new_ip = Helper.ip_to_bytes(generate_random_ip())
    tx = skale.nodes.change_ip(node_id, new_ip, new_ip, wait_for=True, confirmation_blocks=5)
    change_timestamp = skale.nodes.last_change_ip_time(node_id)
    block = skale.web3.eth.get_block(tx.receipt.blockNumber)
    assert change_timestamp == block.timestamp



from skale.utils.account_tools import send_eth
from skale.wallets.web3_wallet import generate_wallet
from tests.manager.delegation.validator_service_test import _generate_new_validator

TEST_RECHARGE_VALUE = 1000
ETH_AMOUNT = 1


def test_get_validator_balance(skale):
    _generate_new_validator(skale)
    latest_id = skale.validator_service.number_of_validators()
    assert skale.wallets.get_validator_balance(latest_id) == 0


def test_recharge_validator_wallet(skale):
    _generate_new_validator(skale)
    latest_id = skale.validator_service.number_of_validators()
    assert skale.wallets.get_validator_balance(latest_id) == 0

    skale.wallets.recharge_validator_wallet(latest_id, value=TEST_RECHARGE_VALUE)
    srw_balance_after = skale.wallets.get_validator_balance(latest_id)
    assert srw_balance_after != 0
    assert srw_balance_after == TEST_RECHARGE_VALUE


def test_withdraw_funds_from_validator_wallet(skale):
    main_wallet = skale.wallet
    new_wallet = generate_wallet(skale.web3)
    send_eth(skale.web3, main_wallet, new_wallet.address, ETH_AMOUNT)
    _generate_new_validator(skale, wallet=new_wallet)

    latest_id = skale.validator_service.number_of_validators()
    skale.wallets.recharge_validator_wallet(latest_id, value=TEST_RECHARGE_VALUE)
    assert skale.wallets.get_validator_balance(latest_id) == TEST_RECHARGE_VALUE

    try:
        skale.wallet = new_wallet
        skale.wallets.withdraw_funds_from_validator_wallet(amount=TEST_RECHARGE_VALUE)
        assert skale.wallets.get_validator_balance(latest_id) == 0
    finally:
        skale.wallet = main_wallet



import pytest
from web3.exceptions import ContractLogicError

from skale.utils.contracts_provision.utils import generate_random_node_data


@pytest.mark.parametrize('number_of_nodes', [1])
def test_get_node(fair, fair_active_nodes):
    registered_node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = registered_node.id

    node = fair.nodes.get(node_id)

    assert node is not None
    assert node.id == node_id
    assert node.address == fair_active_nodes[0].address
    assert isinstance(node.ip_str, str)
    assert isinstance(node.port, int)
    assert node.name == f'node-{node_id}'


@pytest.mark.parametrize('number_of_nodes', [1])
def test_get_public_key(fair, fair_active_nodes):
    registered_node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = registered_node.id
    public_key = fair.nodes.get_public_key(node_id)
    assert isinstance(public_key, str)
    assert public_key.startswith('0x')


@pytest.mark.parametrize('number_of_nodes', [1])
def test_register_active_node(fair, node_wallets):
    main_wallet = fair.wallet
    fair.wallet = node_wallets[0]
    ip, _, port, _ = generate_random_node_data()
    self_stake_requirement = fair.staking.self_stake_requirement()

    with pytest.raises(ContractLogicError):
        fair.nodes.get_by_address(fair.wallet.address)

    active_node_ids_before = fair.nodes.active_node_ids()
    fair.nodes.register_active(ip=ip, port=port, value=self_stake_requirement)
    active_node_ids_after = fair.nodes.active_node_ids()

    assert len(active_node_ids_after) == len(active_node_ids_before) + 1

    node = fair.nodes.get_by_address(fair.wallet.address)
    assert node.address == fair.wallet.address

    fair.wallet = main_wallet


@pytest.mark.parametrize('number_of_nodes', [1])
def test_register_passive_node(fair, node_wallets):
    main_wallet = fair.wallet
    fair.wallet = node_wallets[0]
    ip, _, port, _ = generate_random_node_data()

    with pytest.raises(ContractLogicError):
        fair.nodes.get_passive_node_ids_for_address(fair.wallet.address)

    passive_node_ids_before = fair.nodes.get_passive_node_ids()
    fair.nodes.register_passive(ip=ip, port=port)
    passive_node_ids_after = fair.nodes.get_passive_node_ids()

    assert len(passive_node_ids_after) == len(passive_node_ids_before) + 1

    node_id = fair.nodes.get_passive_node_ids_for_address(fair.wallet.address)[0]
    assert node_id in passive_node_ids_after

    fair.wallet = main_wallet


@pytest.mark.parametrize('number_of_nodes', [1])
def test_set_domain_name(fair, fair_active_nodes):
    main_wallet = fair.wallet
    fair.wallet = fair_active_nodes[0]

    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    domain_name = 'test-domain.example.com'
    fair.nodes.set_domain_name(node_id, domain_name)

    updated_node = fair.nodes.get(node_id)
    assert updated_node.domain_name == domain_name

    fair.wallet = main_wallet


@pytest.mark.parametrize('number_of_nodes', [1])
def test_set_ip_address(fair, fair_active_nodes):
    main_wallet = fair.wallet
    fair.wallet = fair_active_nodes[0]

    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    new_ip, _, new_port, _ = generate_random_node_data()
    fair.nodes.set_ip_address(node_id, new_ip, new_port)

    updated_node = fair.nodes.get(node_id)
    assert updated_node.ip_str == new_ip
    assert updated_node.port == new_port

    fair.wallet = main_wallet


def test_set_committee(fair, node_wallets):
    original_committee_address = fair.committee.address
    committee_address = node_wallets[0].address
    fair.nodes.set_committee(committee_address)

    committee_contract = fair.nodes.contract.functions.committeeContract().call()
    assert committee_contract == committee_address

    fair.nodes.set_committee(original_committee_address)


@pytest.mark.parametrize('number_of_nodes', [2])
def test_request_change_owner(fair, fair_passive_nodes):
    main_wallet = fair.wallet

    node_id = fair.nodes.get_passive_node_ids_for_address(fair_passive_nodes[0].address)[0]
    change_requests_for_node = fair.nodes.contract.functions.ownerChangeRequests(node_id).call()

    assert change_requests_for_node == '0x0000000000000000000000000000000000000000'

    new_owner = fair_passive_nodes[1].address

    fair.wallet = fair_passive_nodes[0]
    fair.nodes.request_change_owner(node_id, new_owner)

    updated_change_requests_for_node = fair.nodes.contract.functions.ownerChangeRequests(
        node_id
    ).call()
    assert updated_change_requests_for_node == new_owner

    fair.wallet = main_wallet


@pytest.mark.parametrize('number_of_nodes', [2])
def test_confirm_owner_change(fair, fair_passive_nodes):
    main_wallet = fair.wallet

    node_id = fair.nodes.get_passive_node_ids_for_address(fair_passive_nodes[0].address)[0]
    new_owner = fair_passive_nodes[1].address

    fair.wallet = fair_passive_nodes[0]
    fair.nodes.request_change_owner(node_id, new_owner)

    change_requests_for_node = fair.nodes.contract.functions.ownerChangeRequests(node_id).call()
    assert change_requests_for_node == new_owner

    fair.wallet = fair_passive_nodes[1]
    fair.nodes.confirm_owner_change(node_id)

    change_requests_for_node_after_confirm = fair.nodes.contract.functions.ownerChangeRequests(
        node_id
    ).call()
    assert change_requests_for_node_after_confirm == '0x0000000000000000000000000000000000000000'

    fair.wallet = main_wallet


@pytest.mark.parametrize('number_of_nodes', [1])
def test_get_id(fair, fair_active_nodes):
    node_id = fair.nodes.get_id(fair_active_nodes[0].address)
    assert node_id > 0

    node = fair.nodes.get(node_id)
    assert node is not None
    assert node.address == fair_active_nodes[0].address


@pytest.mark.parametrize('number_of_nodes', [2])
def test_get_passive_node_ids_for_address(fair, fair_passive_nodes):
    node_ids_first = fair.nodes.get_passive_node_ids_for_address(fair_passive_nodes[0].address)
    node_ids_second = fair.nodes.get_passive_node_ids_for_address(fair_passive_nodes[1].address)
    assert len(node_ids_first) == 1
    assert node_ids_first[0] > 0
    assert len(node_ids_second) == 1
    assert node_ids_second[0] > 0

    main_wallet = fair.wallet
    fair.wallet = fair_passive_nodes[0]
    ip, _, port, _ = generate_random_node_data()

    fair.nodes.register_passive(ip=ip, port=port)

    node_ids_first = fair.nodes.get_passive_node_ids_for_address(fair_passive_nodes[0].address)
    node_ids_second = fair.nodes.get_passive_node_ids_for_address(fair_passive_nodes[1].address)
    assert len(node_ids_first) == 2
    assert node_ids_first[0] > 0
    assert node_ids_first[1] > 0
    assert len(node_ids_second) == 1
    assert node_ids_second[0] > 0

    fair.wallet = main_wallet


@pytest.mark.parametrize('number_of_nodes', [1])
def test_get_passive_node_ids(fair, fair_passive_nodes):
    passive_node_ids = fair.nodes.get_passive_node_ids()

    assert len(passive_node_ids) >= 1
    assert all(node_id > 0 for node_id in passive_node_ids)


@pytest.mark.parametrize('number_of_nodes', [1])
def test_active_node_ids(fair, fair_active_nodes):
    active_node_ids = fair.nodes.active_node_ids()

    assert len(active_node_ids) >= 1
    assert all(node_id > 0 for node_id in active_node_ids)


def test_get_by_address(fair, node_wallets):
    main_wallet = fair.wallet
    fair.wallet = node_wallets[0]
    ip, _, port, _ = generate_random_node_data()
    self_stake_requirement = fair.staking.self_stake_requirement()

    fair.nodes.register_active(ip=ip, port=port, value=self_stake_requirement)
    node = fair.nodes.get_by_address(fair.wallet.address)

    assert node is not None
    assert node.address == fair.wallet.address
    assert node.ip_str == ip
    assert node.port == port

    with pytest.raises(ContractLogicError):
        fair.nodes.get_by_address(node_wallets[1].address)

    fair.wallet = main_wallet


@pytest.mark.parametrize('number_of_nodes', [1])
def test_active_node_exists(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    assert fair.nodes.active_node_exists(node_id) is True
    assert fair.nodes.active_node_exists(999) is False


@pytest.mark.parametrize('number_of_nodes', [1])
def test_passive_node_exists_and_owner_request(fair, fair_passive_nodes):
    node_id = fair.nodes.get_passive_node_ids_for_address(fair_passive_nodes[0].address)[0]
    assert fair.nodes.passive_node_exists(node_id) is True
    assert (
        fair.nodes.get_owner_change_request(node_id) == '0x0000000000000000000000000000000000000000'
    )


def test_committee_contract_accessor(fair):
    assert fair.nodes.committee_contract() == fair.committee.address


def test_decode_public_key(fair):
    raw_public_key = [b'\x12\x34\x56\x78', b'\x9a\xbc\xde\xf0']
    decoded = fair.nodes.decode_public_key(raw_public_key)

    assert isinstance(decoded, str)
    assert decoded.startswith('0x')
    assert len(decoded) == 18


@pytest.mark.parametrize('number_of_nodes', [1])
def test_delete_node(fair, fair_passive_nodes):
    main_wallet = fair.wallet

    node_id = fair.nodes.get_passive_node_ids_for_address(fair_passive_nodes[0].address)[0]

    passive_node_ids_before = fair.nodes.get_passive_node_ids()
    assert node_id in passive_node_ids_before

    fair.wallet = fair_passive_nodes[0]
    fair.nodes.delete_node(node_id)

    passive_node_ids_after = fair.nodes.get_passive_node_ids()
    assert node_id not in passive_node_ids_after
    assert len(passive_node_ids_after) == len(passive_node_ids_before) - 1

    fair.wallet = main_wallet


@pytest.mark.parametrize('number_of_nodes', [1])
def test_delete_node_by_foundation(fair, fair_passive_nodes):
    node_id = fair.nodes.get_passive_node_ids_for_address(fair_passive_nodes[0].address)[0]
    before = fair.nodes.get_passive_node_ids()
    assert node_id in before
    fair.nodes.delete_node_by_foundation(node_id)
    after = fair.nodes.get_passive_node_ids()
    assert node_id not in after



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractEvent, ContractFunction
from web3.types import EventData, Wei

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_contract import SkaleContract
from skale.types.credit_station import PaymentReceivedEvent
from skale.types.schain import SchainName
from skale.utils.helper import schain_name_to_hash

logger = logging.getLogger(__name__)


def get_events_in_chunks(
    event: ContractEvent,
    from_block: int,
    to_block: int | None = None,
    chunk_size: int = 10000,
    **filter_args,
) -> list[EventData]:
    if to_block is None:
        to_block = event.w3.eth.block_number

    events = []
    current_block = from_block
    logger.info(f'Fetching events: {from_block} - {to_block}, chunk_size: {chunk_size}')
    while current_block <= to_block:
        logger.info(
            f'Fetching chunk: {current_block} - {min(current_block + chunk_size - 1, to_block)}'
        )
        end_block = min(current_block + chunk_size - 1, to_block)
        chunk_events = event.get_logs(
            from_block=current_block, to_block=end_block, argument_filters=filter_args
        )
        events.extend(chunk_events)
        current_block = end_block + 1
    return events


class CreditStation(SkaleContract):
    @transaction_method
    def buy(
        self, schain_name: SchainName, purchaser: ChecksumAddress, token: ChecksumAddress
    ) -> ContractFunction:
        return self.contract.functions.buy(schain_name, purchaser, token)

    @transaction_method
    def set_price(self, token: ChecksumAddress, price: Wei) -> ContractFunction:
        return self.contract.functions.setPrice(token, price)

    def get_price(self, token: ChecksumAddress) -> Wei:
        return self.contract.functions.getPrice(token).call()

    def get_prices(self) -> dict[ChecksumAddress, Wei]:
        tokens = self.get_supported_tokens()
        return {token: self.get_price(token) for token in tokens}

    def get_supported_tokens(self) -> list[ChecksumAddress]:
        return self.contract.functions.getSupportedTokens().call()

    def is_token_accepted(self, token: ChecksumAddress) -> bool:
        return self.contract.functions.isTokenAccepted(token).call()

    def get_payment_received_events(
        self,
        from_block: int = 0,
        to_block: int | None = None,
        chunk_size: int = 50000,
        schain_name: SchainName | None = None,
    ) -> list[PaymentReceivedEvent]:
        filter_args = {}
        if schain_name is not None:
            schain_hash = schain_name_to_hash(schain_name)
            filter_args['schainHash'] = schain_hash

        events = get_events_in_chunks(
            self.contract.events.PaymentReceived, from_block, to_block, chunk_size, **filter_args
        )

        return [
            PaymentReceivedEvent(
                payment_id=event['args']['id'],
                schain_hash=event['args']['schainHash'],
                from_address=event['args']['from'],
                to_address=event['args']['to'],
                token_address=event['args']['tokenAddress'],
                block_number=event['blockNumber'],
            )
            for event in events
        ]



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.ima.mainnet.base_deposit_box import BaseDepositBox
from skale.types.schain import SchainName


class DepositBoxERC1155(BaseDepositBox):
    """Class deposit"""

    @transaction_method
    def add_erc1155_token(
        self, schain_name: SchainName, address: ChecksumAddress
    ) -> ContractFunction:
        return self.contract.functions.addERC1155TokenByOwner(schain_name, address)

    @transaction_method
    def deposit_erc1155(
        self, schain_name: SchainName, address: ChecksumAddress, token_id: int, amount: int
    ) -> ContractFunction:
        """Deposit ERC1155"""
        return self.contract.functions.depositERC1155(schain_name, address, token_id, amount)

    @transaction_method
    def deposit_direct_erc1155(
        self,
        schain_name: SchainName,
        address: ChecksumAddress,
        token_id: int,
        amount: int,
        receiver: int,
    ) -> ContractFunction:
        """Direct deposit ERC1155"""
        return self.contract.functions.depositERC1155Direct(
            schain_name, address, token_id, amount, receiver
        )

    @transaction_method
    def deposit_erc1155_batch(
        self, schain_name: SchainName, address: ChecksumAddress, ids: list, amount: list
    ) -> ContractFunction:
        return self.contract.functions.depositERC1155Batch(schain_name, address, ids, amount)

    def get_schain_to_erc1155(
        self, schain_name: SchainName, token_address: ChecksumAddress
    ) -> bool:
        return self.contract.functions.getSchainToERC1155(schain_name, token_address).call()



"""SKALE node rotation test"""

import json
import logging
import os

from skale.contracts.manager.nodes import NodeStatus
from skale.types.dkg import Fp2Point, G2Point, KeyShare
from skale.utils.contracts_provision.fair import set_up_nodes
from skale.utils.contracts_provision.main import _skip_evm_time

logger = logging.getLogger(__name__)


TEST_ROTATION_DELAY = 45260

DIR = os.path.dirname(os.path.realpath(__file__))
DKG_DATA_PATH = os.path.join(DIR, 'dkg_data.json')

with open(DKG_DATA_PATH) as data_file:
    TEST_DKG_DATA = json.loads(data_file.read())


def send_broadcasts(nodes, skale_instances, group_index, skip_node_index=None, rotation_id=0):
    for i, node in enumerate(nodes):
        if i != skip_node_index:
            verification_vector = [
                G2Point(*[Fp2Point(*fp2_point) for fp2_point in g2_point])
                for g2_point in TEST_DKG_DATA['test_verification_vectors'][i]
            ]
            secret_key_contribution = [
                KeyShare(tuple(key_share[0]), key_share[1])
                for key_share in TEST_DKG_DATA['test_encrypted_secret_key_contributions'][i]
            ]
            skale_instances[i].dkg.broadcast(
                group_index,
                node['node_id'],
                verification_vector,
                secret_key_contribution,
                rotation_id,
            )
        else:
            print(f'Skipping broadcast from node {node["node_id"]}')


def send_alrights(nodes, skale_instances, group_index):
    for i, node in enumerate(nodes):
        skale_instances[i].dkg.alright(group_index, node['node_id'])


def send_complaint(nodes, skale_instances, group_index, failed_node_index):
    for i, skale_instance in enumerate(skale_instances):
        if i != failed_node_index:
            failed_node_id = nodes[failed_node_index]['node_id']
            skale_instance.dkg.complaint(group_index, nodes[i]['node_id'], failed_node_id)


def rotate_node(
    skale, group_index, nodes, skale_instances, exiting_node_index, do_dkg=True, rotation_id=0
):
    new_nodes, new_skale_instances = set_up_nodes(skale, 1, no_zero_id=False)
    skale.nodes.init_exit(nodes[exiting_node_index]['node_id'])
    skale_instances[exiting_node_index].manager.node_exit(nodes[exiting_node_index]['node_id'])
    nodes[exiting_node_index] = new_nodes[0]
    skale_instances[exiting_node_index] = new_skale_instances[0]
    if do_dkg:
        run_dkg(nodes, skale_instances, group_index, rotation_id=rotation_id)
    return nodes, skale_instances


def fail_dkg(
    skale,
    nodes,
    skale_instances,
    group_index,
    failed_node_index,
    second_failed_node_index=None,
    rotation_id=0,
) -> list:
    logger.info('Failing first DKG...')
    new_node_ids = []
    new_nodes, new_skale_instances = set_up_nodes(skale, 1, no_zero_id=False)
    new_node_ids.append(new_nodes[0]['node_id'])

    send_broadcasts(nodes, skale_instances, group_index, failed_node_index, rotation_id=rotation_id)
    _skip_evm_time(skale_instances[0].web3, skale.constants_holder.get_dkg_timeout())
    send_complaint(nodes, skale_instances, group_index, failed_node_index)

    nodes[failed_node_index] = new_nodes[0]
    skale_instances[failed_node_index] = new_skale_instances[0]

    if second_failed_node_index:
        logger.info('Failing second DKG...')
        new_nodes, new_skale_instances = set_up_nodes(skale, 1, no_zero_id=False)
        new_node_ids.append(new_nodes[0]['node_id'])

        send_broadcasts(
            nodes,
            skale_instances,
            group_index,
            second_failed_node_index,
            rotation_id=rotation_id + 1,
        )
        _skip_evm_time(skale_instances[0].web3, skale.constants_holder.get_dkg_timeout())
        send_complaint(nodes, skale_instances, group_index, second_failed_node_index)

        nodes[second_failed_node_index] = new_nodes[0]
        skale_instances[second_failed_node_index] = new_skale_instances[0]

    run_dkg(
        nodes,
        skale_instances,
        group_index,
        rotation_id=rotation_id + 1 + (0 if second_failed_node_index is None else 1),
    )
    return new_node_ids


def run_dkg(nodes, skale_instances, group_index, skip_time=True, rotation_id=0):
    logger.info('Running DKG procedure...')
    send_broadcasts(nodes, skale_instances, group_index, rotation_id=rotation_id)
    send_alrights(nodes, skale_instances, group_index)
    if skip_time:
        _skip_evm_time(skale_instances[0].web3, TEST_ROTATION_DELAY)


def remove_node(skale, node_id):
    if skale.nodes.node_status(node_id) == NodeStatus.IN_MAINTENANCE:
        skale.nodes.remove_from_in_maintenance(node_id)
    if skale.nodes.node_status(node_id) != NodeStatus.LEFT:
        skale.nodes.init_exit(node_id)
        skale.manager.node_exit(node_id)



"""SKALE test constants"""

import os
from decimal import Decimal

from eth_typing import HexStr

from skale.types.node import NodeId

DIR_PATH = os.path.dirname(os.path.realpath(__file__))

DEFAULT_NODE_ID = NodeId(1)
DEFAULT_NODE_NAME = 'test_node'
SECOND_NODE_NAME = 'test_node_2'
DEFAULT_NODE_HASH = '23bdf46c41fa300e431425baff124dc31625b34ec09b829f61aa16ab0102ca8d'
DEFAULT_NODE_PORT = 3000

TEST_CONTRACT_NAME = 'Nodes'
TEST_CONTRACT_NAME_HASH = '51468465ffcfb67cf19598bf6dc259b983b35a0ddf02294ef9b1ce0087c47953'

TOKEN_TRANSFER_VALUE = 100
ETH_TRANSFER_VALUE = Decimal('0.05')

DEFAULT_SCHAIN_NAME = 'test_schain_1'
DEFAULT_SCHAIN_INDEX = 0
DEFAULT_SCHAIN_HASH = '0x91bc792a4aa8d4b3ccc3e6acca186f26196ffa90cc391f17a707b829f9479f57'

LIFETIME_YEARS = 1
LIFETIME_SECONDS = LIFETIME_YEARS * 366 * 86400

EMPTY_SCHAIN_ARR = [
    '',
    '0x0000000000000000000000000000000000000000',
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    '0x0000000000000000000000000000000000000000',
]

MIN_NODES_IN_SCHAIN = 2

N_TEST_WALLETS = 2

ENDPOINT = os.environ['ENDPOINT']

HELPER_SCRIPTS_DIR = os.path.join(DIR_PATH, os.pardir, 'helper-scripts')

TEST_ABI_FILEPATH = os.getenv('TEST_ABI_FILEPATH') or os.path.join(
    HELPER_SCRIPTS_DIR, 'contracts_data', 'manager.json'
)
TEST_ALLOCATOR_ABI_FILEPATH = os.getenv('TEST_ALLOCATOR_ABI_FILEPATH') or os.path.join(
    HELPER_SCRIPTS_DIR, 'allocator_contracts_data', 'allocator.json'
)
IMA_DATA_FILEPATH = os.path.join(DIR_PATH, 'ima_data_sample.json')
ETH_PRIVATE_KEY = HexStr(os.environ['ETH_PRIVATE_KEY'])

FAIR_CONTRACTS = os.getenv('FAIR_CONTRACTS')

# constants contract
NEW_REWARD_PERIOD = 600
NEW_DELTA_PERIOD = 400

# sgx wallet
TEST_SGX_ENDPOINT = 'http://localhost:1026'

# rpc wallet
TEST_URL = 'http://localhost:3030'
TEST_RPC_WALLET_URL = 'http://localhost:3000'
NOT_EXISTING_RPC_WALLET_URL = 'http://abc:9999'
EMPTY_HEX_STR = '0x0'

# validator

D_VALIDATOR_ID = 1
D_VALIDATOR_NAME = 'test'
D_VALIDATOR_DESC = 'test'
D_VALIDATOR_FEE = 10
D_VALIDATOR_MIN_DEL = 1000

D_DELEGATION_ID = 0
D_DELEGATION_AMOUNT = 55000000
D_DELEGATION_PERIOD = 2
D_DELEGATION_INFO = 'test'

NOT_EXISTING_ID = NodeId(123123)

MONTH_IN_SECONDS = 60 * 60 * 24 * 31

DELEGATION_STRUCT_LEN = 8

TEST_ECDSA_KEY_NAME = 'NEK:36224eb0296c6c28c3c73942cf28b5ba449e4a1e6472d52d459627c4d9479b21'

TEST_GAS_LIMIT = 10000000

SCHAIN_FIELDS = [
    'name',
    'mainnet_owner',
    'index_in_owner_list',
    'part_of_node',
    'lifetime',
    'start_date',
    'start_block',
    'deposit',
    'index',
    'generation',
    'originator',
    'schain_hash',
    'options',
]



from skale.utils.helper import format_fields, is_valid_ipv4_address


def test_format():
    values = ['value1', 'value2']

    @format_fields(['key1', 'key2'])
    def simple_function():
        return values

    assert simple_function() == {'key1': 'value1', 'key2': 'value2'}

    values = None

    assert simple_function() is None


def test_is_valid_ipv4_address():
    assert is_valid_ipv4_address('1.1.1.1') is True
    assert is_valid_ipv4_address('1.1.1') is False
    assert is_valid_ipv4_address('257.1.1.1') is False
    assert is_valid_ipv4_address('saddas') is False
    assert is_valid_ipv4_address('0:0:0:0:0:0:0:0') is False



from datetime import datetime
from unittest import mock

import pytest
from freezegun import freeze_time

from skale.wallets.redis_wallet import (
    RedisWalletAdapter,
    RedisWalletDroppedError,
    RedisWalletEmptyStatusError,
    RedisWalletNotSentError,
    RedisWalletWaitError,
)
from tests.helper import in_time


@pytest.fixture
def rdp(skale):
    return RedisWalletAdapter(mock.Mock(), 'transactions', skale.wallet)


class RedisTestError(Exception):
    pass


def test_make_raw_id():
    tx_id = RedisWalletAdapter._make_raw_id()
    assert tx_id.startswith(b'tx-')
    assert len(tx_id) == 19


def test_make_score():
    cts = 1623233060
    dt = datetime.utcfromtimestamp(cts)
    with freeze_time(dt):
        score = RedisWalletAdapter._make_score(priority=5)
        assert score == 51623233060


def test_make_record():
    tx = {
        'from': '0x1',
        'to': '0x2',
        'value': 1,
        'gasPrice': 1,
        'gas': 22000,
        'nonce': 1,
        'chainId': 1,
    }
    score = '51623233060'
    tx_id, r = RedisWalletAdapter._make_record(tx, score, 2, method='createNode')
    assert tx_id.startswith(b'tx-') and len(tx_id) == 19
    assert (
        r
        == b'{"status": "PROPOSED", "score": "51623233060", "multiplier": 2, "tx_hash": null, "method": "createNode", "from": "0x1", "to": "0x2", "value": 1, "gasPrice": 1, "gas": 22000, "nonce": 1, "chainId": 1}'  # noqa
    )


def test_sign_and_send(rdp):
    tx = {
        'from': '0x1',
        'to': '0x2',
        'value': 1,
        'gasPrice': 1,
        'gas': 22000,
        'nonce': 1,
        'chainId': 1,
    }
    tx_id = rdp.sign_and_send(tx, multiplier=2, priority=5)
    assert str(tx_id).startswith('tx-') and len(tx_id) == 19

    rdp.rs.pipeline = mock.Mock(side_effect=RedisTestError('rtest'))
    with pytest.raises(RedisWalletNotSentError):
        tx_id = rdp.sign_and_send(tx, multiplier=2, priority=5)


def test_rdp_wait(rdp):
    tx_id = 'tx-tttttttttttttttt'
    rdp.get_record = mock.Mock(return_value=None)
    with in_time(3):
        with pytest.raises(RedisWalletEmptyStatusError):
            rdp.wait(tx_id, timeout=2)

    rdp.get_record = mock.Mock(return_value={'tx_hash': 'test', 'status': 'DROPPED'})
    with in_time(2):
        with pytest.raises(RedisWalletDroppedError):
            rdp.wait(tx_id, timeout=100)

    rdp.get_record = mock.Mock(side_effect=RedisTestError('test'))
    with in_time(2):
        with pytest.raises(RedisWalletWaitError):
            rdp.wait(tx_id, timeout=100)

    rdp.get_record = mock.Mock(return_value={'tx_hash': 'test', 'status': 'SUCCESS'})
    fake_receipt = {'test': 'test'}
    with mock.patch('skale.wallets.redis_wallet.get_receipt', return_value=fake_receipt):
        with in_time(2):
            assert rdp.wait(tx_id, timeout=100) == fake_receipt



from unittest import mock
from unittest.mock import Mock

import web3
from hexbytes import HexBytes

from skale.contracts.manager.dkg import G2Point, KeyShare
from skale.utils.helper import schain_name_to_hash, split_public_key

SCHAIN_NAME = 'pointed-asellus-australis'
PUBLIC_KEY = '0xfcb3765bdb954ab0672fce731583ad8a94cf05fe63c147f881f8feea18e072d4cad3ec142a65de66a1d50e4fc34a7841c5488ccb55d02cf86013208c17517d64'  # noqa


def assert_transaction_data_in_mock(expected_txn, send_tx_mock):
    assert expected_txn['data'][2:] in HexBytes(send_tx_mock.call_args[0][0]).hex()


def test_response(skale):
    nonce = skale.web3.eth.get_transaction_count(skale.wallet.address)
    contract_address = skale.dkg.address
    chain_id = skale.web3.eth.chain_id
    gas_limit = 11000000
    expected_txn = {
        'value': 0,
        'gasPrice': skale.web3.eth.gas_price,
        'chainId': chain_id,
        'gas': gas_limit,
        'nonce': nonce,
        'type': 1,
        'to': contract_address,
        'data': (
            '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'  # noqa
        ),
    }
    from_node_index = 0
    group_index = schain_name_to_hash(SCHAIN_NAME)
    share = group_index  # not an invariant, only a mock
    secret_number = 1
    multiplied_share = G2Point((1, 2), (3, 4))
    verification_vector = [G2Point((1, 2), (3, 4)) for i in range(0, 3)]
    verification_vector_mult = [G2Point((1, 2), (3, 4)) for i in range(0, 3)]
    secret_key_contribution = [KeyShare(split_public_key(PUBLIC_KEY), share)]

    with mock.patch.object(
        skale.dkg.contract.functions.preResponse, 'call', new=Mock(return_value=[])
    ):
        with mock.patch.object(web3.eth.Eth, 'send_raw_transaction') as send_tx_mock:
            send_tx_mock.return_value = b'hexstring'
            skale.dkg.pre_response(
                group_index=group_index,
                from_node_index=from_node_index,
                verification_vector=verification_vector,
                verification_vector_mult=verification_vector_mult,
                secret_key_contribution=secret_key_contribution,
                wait_for=False,
                gas_limit=gas_limit,
            )
            send_tx_mock.assert_called_with(mock.ANY)
            assert_transaction_data_in_mock(expected_txn, send_tx_mock)

    expected_txn = {
        'value': 0,
        'gasPrice': skale.web3.eth.gas_price,
        'chainId': chain_id,
        'gas': gas_limit,
        'nonce': nonce,
        'to': contract_address,
        'type': 1,
        'data': (
            '0x6ef89763e332bac19e758fe13db6129827da76846b8c6d26f1e70385d3f0afc0299e3db9000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000010000000000000000000000000000000000000000000000000000000000000001000000000000000000000000000000000000000000000000000000000000000200000000000000000000000000000000000000000000000000000000000000030000000000000000000000000000000000000000000000000000000000000004'  # noqa
        ),
    }

    with mock.patch.object(
        skale.dkg.contract.functions.response, 'call', new=Mock(return_value=[])
    ):
        with mock.patch.object(web3.eth.Eth, 'send_raw_transaction') as send_tx_mock:
            send_tx_mock.return_value = b'hexstring'
            skale.dkg.response(
                group_index=group_index,
                from_node_index=from_node_index,
                secret_number=secret_number,
                multiplied_share=multiplied_share,
                wait_for=False,
                gas_limit=gas_limit,
            )
            send_tx_mock.assert_called_with(mock.ANY)
            assert_transaction_data_in_mock(expected_txn, send_tx_mock)


def test_alright(skale):
    nonce = skale.web3.eth.get_transaction_count(skale.wallet.address)
    contract_address = skale.dkg.address
    chain_id = skale.web3.eth.chain_id
    gas_limit = 10000000
    expected_txn = {
        'value': 0,
        'gasPrice': skale.web3.eth.gas_price,
        'chainId': chain_id,
        'gas': gas_limit,
        'nonce': nonce,
        'type': 1,
        'to': contract_address,
        'data': (
            '0xb9799682e332bac19e758fe13db6129827da76846b8c6d26f1e70385d3f0afc0299e3db90000000000000000000000000000000000000000000000000000000000000000'  # noqa
        ),
    }
    group_index = schain_name_to_hash(SCHAIN_NAME)
    from_node_index = 0

    with mock.patch.object(skale.dkg.contract.functions.alright, 'call', new=Mock(return_value=[])):
        with mock.patch.object(web3.eth.Eth, 'send_raw_transaction') as send_tx_mock:
            send_tx_mock.return_value = b'hexstring'
            skale.dkg.alright(group_index, from_node_index, wait_for=False, gas_limit=gas_limit)
            assert_transaction_data_in_mock(expected_txn, send_tx_mock)


def test_complaint(skale):
    nonce = skale.web3.eth.get_transaction_count(skale.wallet.address)
    contract_address = skale.dkg.address
    chain_id = skale.web3.eth.chain_id
    gas_limit = 8000000
    expected_txn = {
        'value': 0,
        'gasPrice': skale.web3.eth.gas_price,
        'chainId': chain_id,
        'gas': gas_limit,
        'nonce': nonce,
        'to': contract_address,
        'type': 1,
        'data': (
            '0xd76c2c4fe332bac19e758fe13db6129827da76846b8c6d26f1e70385d3f0afc0299e3db900000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'  # noqa
        ),
    }
    group_index = schain_name_to_hash(SCHAIN_NAME)
    from_node_index = 0
    to_node_index = 0

    with mock.patch.object(
        skale.dkg.contract.functions.complaint, 'call', new=Mock(return_value=[])
    ):
        with mock.patch.object(web3.eth.Eth, 'send_raw_transaction') as send_tx_mock:
            send_tx_mock.return_value = b'hexstring'
            skale.dkg.complaint(
                group_index, from_node_index, to_node_index, wait_for=False, gas_limit=gas_limit
            )
            assert_transaction_data_in_mock(expected_txn, send_tx_mock)


def test_complaint_bad_data(skale):
    nonce = skale.web3.eth.get_transaction_count(skale.wallet.address)
    contract_address = skale.dkg.address
    chain_id = skale.web3.eth.chain_id
    gas_limit = 8000000
    expected_txn = {
        'value': 0,
        'gasPrice': skale.web3.eth.gas_price,
        'chainId': chain_id,
        'gas': gas_limit,
        'nonce': nonce,
        'to': contract_address,
        'type': 1,
        'data': (
            '0xd76c2c4fe332bac19e758fe13db6129827da76846b8c6d26f1e70385d3f0afc0299e3db900000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'  # noqa
        ),
    }
    group_index = schain_name_to_hash(SCHAIN_NAME)
    from_node_index = 0
    to_node_index = 0

    with mock.patch.object(
        skale.dkg.contract.functions.complaintBadData, 'call', new=Mock(return_value=[])
    ):
        with mock.patch.object(web3.eth.Eth, 'send_raw_transaction') as send_tx_mock:
            send_tx_mock.return_value = b'hexstring'
            skale.dkg.complaint(
                group_index, from_node_index, to_node_index, wait_for=False, gas_limit=gas_limit
            )
            assert_transaction_data_in_mock(expected_txn, send_tx_mock)


def test_is_last_dkg_successful(skale, schain):
    group_index = skale.web3.keccak(text=schain)
    assert not skale.dkg.is_last_dkg_successful(group_index)


def test_channel_opened(skale, schain):
    group_index = skale.web3.keccak(text=schain)
    assert skale.dkg.is_channel_opened(group_index)


def test_broadcast_possible(skale, nodes, schain, node_wallets):
    group_index = skale.web3.keccak(text=schain)
    node_id, *_ = nodes
    wallet, *_ = node_wallets
    assert skale.dkg.is_broadcast_possible(group_index, node_id, wallet.address)


def test_alright_possible(skale, nodes, schain, node_wallets):
    group_index = skale.web3.keccak(text=schain)
    node_id, *_ = nodes
    wallet, *_ = node_wallets
    assert not skale.dkg.is_alright_possible(group_index, node_id, wallet.address)


def test_pre_response_possible(skale, nodes, schain, node_wallets):
    group_index = skale.web3.keccak(text=schain)
    node_id, *_ = nodes
    wallet, *_ = node_wallets
    assert not skale.dkg.is_pre_response_possible(group_index, node_id, wallet.address)


def test_response_possible(skale, nodes, schain, node_wallets):
    group_index = skale.web3.keccak(text=schain)
    node_id, *_ = nodes
    wallet, *_ = node_wallets
    assert not skale.dkg.is_response_possible(group_index, node_id, wallet.address)


def test_everyone_broadcasted(skale, schain):
    group_index = skale.web3.keccak(text=schain)
    assert not skale.dkg.is_everyone_broadcasted(group_index, skale.wallet.address)


def test_number_of_completed(skale, schain):
    group_index = skale.web3.keccak(text=schain)
    assert skale.dkg.get_number_of_completed(group_index) == 0


def test_channel_started_time(skale, schain):
    group_index = skale.web3.keccak(text=schain)
    assert skale.dkg.get_channel_started_time(group_index) != 0


def test_complaint_started_time(skale, schain):
    group_index = skale.web3.keccak(text=schain)
    assert skale.dkg.get_complaint_started_time(group_index) == 0


def test_alright_started_time(skale, schain):
    group_index = skale.web3.keccak(text=schain)
    assert skale.dkg.get_alright_started_time(group_index) == 0


def test_get_complaint_data(skale, schain):
    group_index = skale.web3.keccak(text=schain)
    assert skale.dkg.get_complaint_data(group_index) != [0, 0]


def test_time_of_last_successful_dkg(skale, schain):
    group_index = skale.web3.keccak(text=schain)
    assert skale.dkg.get_time_of_last_successful_dkg(group_index) == 0



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2025-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import functools
import logging
import math

from skale.contracts.base_contract import transaction_method
from skale.contracts.fair.nodes import Nodes
from skale.contracts.skale_contract import SkaleContract
from skale.types.committee import Timestamp
from skale.types.node import NodeId

logger = logging.getLogger(__name__)


class Status(SkaleContract):
    @property
    @functools.lru_cache()
    def nodes(self) -> 'Nodes':
        return self.skale.nodes

    def last_heartbeat_timestamp(self, node_id: NodeId) -> Timestamp:
        return self.contract.functions.lastHeartbeatTimestamp(node_id).call()

    def heartbeat_interval(self) -> int:
        return self.contract.functions.heartbeatInterval().call()

    def is_healthy(self, node_id: NodeId) -> bool:
        return self.contract.functions.isHealthy(node_id).call()

    def get_whitelisted_nodes(self) -> list[NodeId]:
        return self.contract.functions.getWhitelistedNodes().call()

    def is_whitelisted(self, node_id: NodeId) -> bool:
        return self.contract.functions.isWhitelisted(node_id).call()

    def active_whitelisted_node_ids(self) -> list[NodeId]:
        return list(set(self.nodes.active_node_ids()) & set(self.get_whitelisted_nodes()))

    def calc_alive_gas_limit(self) -> int:
        active_whitelisted_nodes = len(self.active_whitelisted_node_ids())
        alive_gas_limit_float = (230000 * math.log10(active_whitelisted_nodes + 15) + 720000) * 1.2
        alive_gas_limit = int(alive_gas_limit_float)
        logger.info(
            'alive_gas_limit: %s, active_whitelisted_nodes: %s',
            alive_gas_limit,
            active_whitelisted_nodes,
        )
        return alive_gas_limit

    @transaction_method
    def alive(self):
        return self.contract.functions.alive()

    @transaction_method
    def set_heartbeat_interval(self, interval: int):
        return self.contract.functions.setHeartbeatInterval(interval)

    @transaction_method
    def whitelist_node(self, node_id: NodeId):
        return self.contract.functions.whitelistNode(node_id)

    @transaction_method
    def remove_node_from_whitelist(self, node_id: NodeId):
        return self.contract.functions.removeNodeFromWhitelist(node_id)



# flake8: noqa
from skale.contracts.skale_contract import SkaleContract
from skale.contracts.base_contract import transaction_method
from skale.contracts.etherbase.etherbase import Etherbase

__all__ = ['Etherbase']



import sys

if sys.version_info < (3, 11):
    raise EnvironmentError('Python 3.11 or above is required')

from skale.fair_manager import FairManager
from skale.mainnet_credit_station import MainnetCreditStation
from skale.mainnet_ima import MainnetIma
from skale.mainnet_ima import MainnetIma as SkaleIma  # todo: deprecated, to be removed in v8
from skale.schain_credit_station import SchainCreditStation
from skale.schain_ima import SchainIma
from skale.skale_allocator import SkaleAllocator
from skale.skale_manager import SkaleManager

__all__ = [
    'SkaleManager',
    'SkaleAllocator',
    'SkaleIma',
    'MainnetIma',
    'FairManager',
    'SchainIma',
    'MainnetCreditStation',
    'SchainCreditStation',
]



"""Tests for contracts/delegation/distributor.py"""

from tests.constants import D_VALIDATOR_ID


def test_get_earned_bounty_amount(skale):
    res = skale.distributor.get_earned_bounty_amount(D_VALIDATOR_ID, skale.wallet.address)
    assert res['earned'] == 0


def test_get_earned_fee_amount(skale):
    res = skale.distributor.get_earned_fee_amount(skale.wallet.address)
    assert res['earned'] == 0


# todo: improve test
def test_withdraw_bounty(skale):
    tx_res = skale.distributor.withdraw_bounty(
        validator_id=D_VALIDATOR_ID, to=skale.wallet.address, wait_for=True
    )
    assert tx_res.receipt['status'] == 1


# todo: improve test
def test_withdraw_fee(skale):
    tx_res = skale.distributor.withdraw_fee(to=skale.wallet.address, wait_for=True)
    assert tx_res.receipt['status'] == 1
